import itertools

from bygg.cmd.datastructures import ByggContext, SubProcessIpcDataTree, get_entrypoints
from bygg.output.output import TerminalStyle as TS


class TreeStyle:
    """Base class for tree styles."""

    BAR = "─"
    PIPE = "│"
    T_JOINT = "├"
    END_CORNER = "└"
    CONNECTOR = T_JOINT + BAR * 2 + " "
    HANGER = END_CORNER + BAR * 2 + " "


class TreeStyleUnicode(TreeStyle):
    """Same style as base class, but with parametrized indent."""

    def __init__(self, indent=4):
        self.CONNECTOR = (
            f"{self.T_JOINT + self.BAR * (indent - 1 - len(self.T_JOINT)):<{indent}}"
        )
        self.HANGER = f"{self.END_CORNER + self.BAR * (indent - 1 - len(self.END_CORNER)):<{indent}}"


class TreeStyleAscii(TreeStyle):
    """ASCII characters instead of unicode."""

    BAR = "-"
    PIPE = "|"
    T_JOINT = "+"
    END_CORNER = "\\"

    def __init__(self, indent=4):
        self.CONNECTOR = (
            f"{self.T_JOINT + self.BAR * (indent - 1 - len(self.T_JOINT)):<{indent}}"
        )
        self.HANGER = f"{self.END_CORNER + self.BAR * (indent - 1 - len(self.END_CORNER)):<{indent}}"


def print_tree(ipc_data_tree: SubProcessIpcDataTree, actions: list[str]):
    """Print the dependency tree from the IPC data."""
    actions_to_list = actions if actions else sorted(ipc_data_tree.actions)
    trees = [
        x for x in [ipc_data_tree.actions.get(a, "") for a in actions_to_list] if x
    ]
    if trees:
        # Python 3.11 f-strings don't support `\n`
        print("\n" + "\n".join(trees))


def tree_collect_for_environment(
    ctx: ByggContext, environment_name: str
) -> SubProcessIpcDataTree:
    """Collect the currently loaded entrypoints and render their respective dependency
    trees.

    Example output:

    all_checks
    ├── check_inputs_outputs
    │   └── circular_C
    │       └── circular_B
    │           └── circular_A
    └── output_file_missing
        └── no_outputs_A
    """

    entrypoints = get_entrypoints(ctx, environment_name)

    indent = 4
    style = TreeStyleUnicode(indent)

    formatted_data: dict[str, str] = {}

    for entrypoint in entrypoints:

        def format_children(name: str, last_sibling: bool, depth: int) -> list[str]:
            action = ctx.scheduler.build_actions[name]
            display_name = f"{TS.BOLD}{name}{TS.RESET}" if depth == 0 else name

            # Format children and flatten:
            children = itertools.chain.from_iterable(
                (
                    format_children(dep, i == len(action.dependencies) - 1, depth + 1)
                    for i, dep in enumerate(sorted(action.dependencies))
                )
            )

            # Indent with the correct prefixes:
            prefix = style.HANGER if last_sibling else style.CONNECTOR
            subtree = [f"{prefix if depth > 0 else ''}{display_name}"]

            child_prefix = f"{style.PIPE if not last_sibling else ' ':<{indent}}"
            subtree.extend(
                [f"{child_prefix if depth > 0 else ''}{item}" for item in children]
            )
            return subtree

        formatted_data[entrypoint.name] = "\n".join(
            format_children(entrypoint.name, True, 0)
        )

    return SubProcessIpcDataTree(actions=formatted_data)



from dataclasses import dataclass
from typing import Literal

RunnerInstruction = Literal["restart_build", "exit_job_failed"]


@dataclass
class CommandStatus:
    """The status of a command."""

    rc: int  # return code; follows shell conventions where 0 is success
    message: str | None = None  # a message to display to the user
    output: str | None = None  # output of the command
    runner_instruction: RunnerInstruction | None = None  # instruction to the runner


JobStatus = Literal["queued", "running", "finished", "failed", "stopped", "skipped"]

Severity = Literal["error", "warning", "info"]



import argparse
from collections.abc import Callable
import contextlib
import os
import pickle
import subprocess
import sys
import tempfile
from typing import Optional, TypeAlias

from bygg.cmd.argument_parsing import ByggNamespace, create_argument_parser
from bygg.cmd.argument_unparsing import unparse_args
from bygg.cmd.completions import (
    do_completion,
    generate_shell_completions,
    is_completing,
)
from bygg.cmd.configuration import (
    DEFAULT_ENVIRONMENT_NAME,
    Byggfile,
    dump_schema,
    has_byggfile,
    read_config_files,
)
from bygg.cmd.datastructures import (
    ByggContext,
    SubProcessIpcData,
    SubProcessIpcDataTree,
    get_entrypoints,
)
from bygg.cmd.environments import (
    get_environment_for_action,
    load_environment,
    setup_environment,
    should_restart_with,
)
from bygg.cmd.list_actions import list_collect_for_environment, print_actions
from bygg.cmd.maintenance import perform_maintenance
from bygg.cmd.tree import print_tree, tree_collect_for_environment
from bygg.core.runner import ProcessRunner
from bygg.core.scheduler import Scheduler
from bygg.logutils import logger
from bygg.output.output import TerminalStyle as TS
from bygg.output.output import output_error, output_info, output_warning
from bygg.output.status_display import (
    get_on_job_status,
    on_runner_status,
)
from bygg.system_helpers import change_dir


def init_bygg_context(
    configuration: Byggfile,
    parser: argparse.ArgumentParser,
    args_namespace: argparse.Namespace,
    args: ByggNamespace,
) -> ByggContext:
    scheduler = Scheduler()
    runner = ProcessRunner(scheduler)

    # Set up status listeners
    runner.job_status_listener = get_on_job_status(args, configuration)
    runner.runner_status_listener = on_runner_status

    return ByggContext(
        runner,
        scheduler,
        configuration,
        parser,
        args_namespace,
        args,
        SubProcessIpcData(),
    )


def print_version():
    import importlib.metadata

    output_info(f"bygg {importlib.metadata.version('bygg')}")


DISPATCHER_IS_COMPLETING_EXIT_CODE = 126
"""Exit code returned by a subprocess when completing."""

DISPATCHER_ACTION_NOT_FOUND_EXIT_CODE = 127
"""Exit code returned by a subprocess when an action is not found."""


def bygg():
    """Entry point for the Bygg command line interface."""
    parser = create_argument_parser()
    do_completion(parser)
    args = parser.parse_args()
    if not args.is_restarted_with_env:
        # Build and potentially watch:
        rc = 0
        while True:
            with change_dir(None):  # change back to the starting dir
                environment_data_list = parent_dispatcher(parser, args)

                rc = output_status_codes(environment_data_list)
                files_to_watch = extract_input_files(environment_data_list)
                if not args.watch or len(files_to_watch) == 0:
                    break

                from bygg.cmd.watch import do_watch

                output_info("Watching for changes")
                do_watch(files_to_watch)
        return rc
    else:
        subprocess_dispatcher(parser, args)


def output_status_codes(
    environment_data_list: list[dict[str, SubProcessIpcData]],
) -> int:
    rc = 0
    error_lines: list[str] = []
    for environment_data in environment_data_list:
        for env in environment_data.values():
            for name, command_status in env.failed_jobs.items():
                if command_status.rc != 0:
                    error_lines.append(f"{name} [{command_status.rc}]")
                    rc = command_status.rc
    if len(error_lines):
        plural_s = "s" if len(error_lines) > 1 else ""
        output_error(
            f"The following action{plural_s} exited with a non-zero status code:"
        )
        for line in error_lines:
            output_error(f"- {line}")

    return rc


def extract_input_files(
    environment_data_list: list[dict[str, SubProcessIpcData]],
) -> set[str]:
    files_to_watch: set[str] = set()
    for environment_data in environment_data_list:
        for env in environment_data.values():
            files_to_watch.update(env.found_input_files)
    return files_to_watch


def parent_dispatcher(
    parser: argparse.ArgumentParser,
    args_namespace: argparse.Namespace,
) -> list[dict[str, SubProcessIpcData]]:
    """
    Takes both argparse.ArgumentParser and argparse.Namespace arguments since it can be
    called also from completers. However, it is not used by subprocesses.
    """

    args = ByggNamespace(**vars(args_namespace))
    assert not args.is_restarted_with_env

    # Generate completions, dump schema or print the version
    generate_shell_completions(args.completions)

    if args.dump_schema:
        dump_schema()
        sys.exit(0)

    if args.version:
        print_version()
        sys.exit(0)

    # Change directory if needed
    directory = args.directory[0] if args.directory else None

    if directory:
        directory_arg = args.directory[0]
        output_info(f"Entering directory '{directory_arg}'")
        os.chdir(directory_arg)

    if args.watch and args.maintenance_commands is not None:
        # The other build setup group commands are blocked by argparse
        output_error("Cannot use -w/--watch with maintenance commands")
        sys.exit(1)

    # No byggfiles
    if not has_byggfile():
        output_error("No build files found.")
        sys.exit(1)

    # Read static configuration
    configuration = read_config_files()

    # Perform maintenance commands
    if args.maintenance_commands:
        perform_maintenance(configuration, args.maintenance_commands)
        sys.exit(0)

    # Create runner and scheduler and such
    ctx = init_bygg_context(configuration, parser, args_namespace, args)

    actions_to_build = [*args.actions]
    if not args.actions and ctx.configuration.settings.default_action is not None:
        actions_to_build.append(ctx.configuration.settings.default_action)

    logger.info("Actions to be built: %s", actions_to_build)

    only_collect = (
        not actions_to_build or args.list_actions or args.tree or is_completing()
    )

    # We have nothing to build, but other things to do
    if only_collect:
        environment_data = do_in_all_environments(ctx, run_or_collect_in_environment)

        if is_completing():
            return [environment_data]

        if args.list_actions:
            print_actions(ctx, environment_data)
            sys.exit(0)

        if args.tree:
            if not actions_to_build:
                output_error(
                    "The --tree command requires at least one action to be specified."
                )
                output_error(
                    "No actions were specified and no default action is defined."
                )
                print_actions(ctx, environment_data)
                sys.exit(1)
            tree_actions = set(actions_to_build)
            for v in environment_data.values():
                if v.tree:
                    tree_actions = tree_actions - v.found_actions
                    print_tree(v.tree, actions_to_build)
            if tree_actions:
                not_found_actions = [
                    f"'{a}'"
                    for a in (filter(lambda x: x in tree_actions, actions_to_build))
                ]
                output_error(
                    f"Error: The following action{'s' if len(not_found_actions) > 1 else ''} could not be found: {TS.BOLD}{', '.join(not_found_actions)}{TS.NOBOLD}."
                )
                sys.exit(1)
            sys.exit(0)

        print_actions(ctx, environment_data)
        sys.exit(1)

    # Here we have something to build

    environment_data_list = []

    for action in actions_to_build:
        environment_data = do_in_all_environments(
            ctx,
            lambda ctx, environment_name: run_or_collect_in_environment(
                ctx, environment_name, action=action
            ),
            start_with_env=get_environment_for_action(ctx, action),
        )
        all_found_actions = set().union(
            *(env.found_actions for env in environment_data.values())
        )

        if action not in all_found_actions:
            output_error(
                f"Error: The following action could not be found: {TS.BOLD}'{action}'{TS.NOBOLD}."
            )
            sys.exit(1)

        environment_data_list.append(environment_data)

    return environment_data_list


DoerType: TypeAlias = Callable[[ByggContext, str], tuple[SubProcessIpcData, bool]]
"""Takes a ByggContext and an action name as arguments and returns a tuple of the
subprocess IPC data and whether the action was handled."""


def do_in_all_environments(
    ctx: ByggContext,
    doer: DoerType,
    *,
    start_with_env: Optional[str] = None,
) -> dict[str, SubProcessIpcData]:
    """
    Runs doer for all environments and returns the environment data.
    Starts with start_with_env if specified.
    """
    environment_data: dict[str, SubProcessIpcData] = {}
    environment_names = [DEFAULT_ENVIRONMENT_NAME, *ctx.configuration.environments]
    if start_with_env:
        environment_names.remove(start_with_env)
        environment_names.insert(0, start_with_env)

    for environment_name in environment_names:
        environment_data[environment_name], early_out = doer(ctx, environment_name)
        if early_out:
            break
    return environment_data


def run_or_collect_in_environment(
    ctx: ByggContext,
    environment_name: str,
    *,
    action: Optional[str] = None,
) -> tuple[SubProcessIpcData, bool]:
    """
    Runs the given action in the given environment or just collects data about the
    environment. These two flows are roughly the same.
    """
    subprocess_data = SubProcessIpcData()

    if environment_name != DEFAULT_ENVIRONMENT_NAME:
        # Set up the environment if we should not run in the ambient one. Spawn
        # subprocess if necessary.
        logger.info("Running-collecting: setting up environment %s", environment_name)
        environment = ctx.configuration.environments.get(environment_name, None)
        assert environment
        setup_environment(ctx, environment)
        subprocess_bygg_path = should_restart_with(environment)

        logger.info("Running action: should restart with %s", subprocess_bygg_path)
        if subprocess_bygg_path:
            result = spawn_subprocess(
                ctx,
                environment_name=environment_name,
                subprocess_bygg_path=subprocess_bygg_path,
                action=action,
            )
            action_found = action is not None and action in result.found_actions
            return (result, action_found)

    logger.info("Running-collecting: in ambient environment")
    load_environment(ctx, environment_name)
    logger.debug(
        "Entrypoints found for %s: %s",
        environment_name,
        get_entrypoints(ctx, environment_name),
    )
    subprocess_data.found_actions = {
        e.name for e in get_entrypoints(ctx, environment_name)
    }
    subprocess_data.list = list_collect_for_environment(ctx, environment_name)
    subprocess_data.tree = (
        # Only collect tree data if not completing, since the completion tester in
        # pytest messes with code that is loaded dynamically in examples/trivial so that
        # tree doesn't work. Might be fixable, but that's for future Homer. Completion
        # code works fine when called interactively and not from pytest.
        tree_collect_for_environment(ctx, environment_name)
        if not is_completing()
        else SubProcessIpcDataTree({})
    )

    # Our work here is done, or, we had nothing to do
    if action is None or action not in subprocess_data.found_actions:
        return (subprocess_data, False)

    # Build or clean handled below. These are the only commands handled here.
    from bygg.cmd.build_clean import build, clean

    if ctx.bygg_namespace.clean:
        status = clean(ctx, action)
    else:
        status, input_files = build(
            ctx,
            action,
            ctx.bygg_namespace.jobs,
            ctx.bygg_namespace.always_make,
            ctx.bygg_namespace.check,
        )
        subprocess_data.found_input_files.update(input_files)
    if not status:
        subprocess_data.failed_jobs = {
            job.name: job.status
            for job in ctx.runner.failed_jobs
            if job.status is not None
        }

    # All critical errors will already have done sys.exit
    return (subprocess_data, True)


@contextlib.contextmanager
def auto_tmpfile():
    ipc_filename: str | None = None
    try:
        # Create IPC file
        fd, ipc_filename = tempfile.mkstemp()
        os.close(fd)
        yield ipc_filename
    finally:
        if ipc_filename:
            os.remove(ipc_filename)


def spawn_subprocess(
    ctx: ByggContext,
    *,
    environment_name: str,
    subprocess_bygg_path: str,
    action: str | None = None,
) -> SubProcessIpcData:
    # Create IPC file

    with auto_tmpfile() as ipc_filename:
        exec_list = []
        exec_list += [subprocess_bygg_path]
        if action:
            exec_list += [action]
        exec_list += unparse_args(
            ctx.parser,
            ctx.args_namespace,
            drop=["actions", "maintenance_commands", "watch"],
        )
        exec_list += ["--is_restarted_with_env", environment_name]
        exec_list += ["--ipc_filename", ipc_filename]

        logger.debug("Restarting with: %s", exec_list)
        return_code = 0
        try:
            subprocess.run(exec_list, encoding="utf-8", check=True)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            match e:
                case subprocess.CalledProcessError() if (
                    e.returncode == DISPATCHER_IS_COMPLETING_EXIT_CODE
                    or e.returncode == DISPATCHER_ACTION_NOT_FOUND_EXIT_CODE
                ):
                    # NOP
                    logger.info("Subprocess returned with code %s", e.returncode)
                    return_code = e.returncode
                    pass
                case subprocess.CalledProcessError():
                    output_error(
                        f"Action {TS.BOLD}{action}{TS.NOBOLD}: status {e.returncode}"
                    )
                    sys.exit(e.returncode)
                case FileNotFoundError():
                    output_error(
                        f"Error: Could not restart with '{subprocess_bygg_path}'."
                    )
                    output_warning(
                        "Make sure that bygg is in your pip requirements list for this environment."
                    )
                    sys.exit(1)

        try:
            logger.info("Reading IPC file '%s'", ipc_filename)
            with open(ipc_filename, "rb") as f:
                subprocess_data: SubProcessIpcData = pickle.load(f)
                subprocess_data.return_code = return_code
                return subprocess_data
        except (FileNotFoundError, pickle.PickleError) as e:
            logger.error("Error: %s, %s", e.__class__, e)
            match e:
                case FileNotFoundError():
                    output_error(f"Error: Could not open IPC file '{ipc_filename}'.")
                    sys.exit(1)
                case pickle.PickleError():
                    output_error(f"Error: Invalid IPC file '{ipc_filename}'.")
                    sys.exit(1)

    assert not "this point should not be reached"
    sys.exit(1)


def subprocess_dispatcher(parser, args_namespace):
    # We're in subprocess

    # Always called with at most one action at a time; it is up to the caller to loop
    # over the action list.

    args = ByggNamespace(**vars(args_namespace))
    assert args.is_restarted_with_env
    environment_name = args.is_restarted_with_env
    logger.info(
        "Running action '%s' in environment '%s'.",
        args.actions,
        environment_name,
    )

    # Read static configuration
    configuration = read_config_files()

    # Create runner and scheduler and such
    ctx = init_bygg_context(configuration, parser, args_namespace, args)

    ctx.ipc_data = SubProcessIpcData()
    load_environment(ctx, environment_name)

    logger.debug(
        "Entrypoints found for %s: %s",
        environment_name,
        get_entrypoints(ctx, environment_name),
    )
    ctx.ipc_data.found_actions = {
        e.name for e in get_entrypoints(ctx, environment_name)
    }
    action = args.actions[0] if args.actions else None

    ctx.ipc_data.list = list_collect_for_environment(ctx, environment_name)
    ctx.ipc_data.tree = tree_collect_for_environment(ctx, environment_name)

    if is_completing():
        write_ipc_data(ctx, args)
        sys.exit(DISPATCHER_IS_COMPLETING_EXIT_CODE)

    if action and action not in ctx.scheduler.build_actions or not action:
        write_ipc_data(ctx, args)
        sys.exit(DISPATCHER_ACTION_NOT_FOUND_EXIT_CODE)

    from bygg.cmd.build_clean import build, clean

    if args.clean:
        status = clean(ctx, action)
    else:
        status, input_files = build(
            ctx, action, args.jobs, args.always_make, args.check
        )
        ctx.ipc_data.found_input_files.update(input_files)

    if not status:
        sys.exit(1)

    write_ipc_data(ctx, args)
    sys.exit(0)


def write_ipc_data(ctx: ByggContext, args: ByggNamespace):
    ipc_filename = args.ipc_filename[0] if args.ipc_filename else None
    if ipc_filename:
        logger.debug("Writing IPC data to %s", args.ipc_filename)
        with open(ipc_filename, "wb") as f:
            pickle.dump(ctx.ipc_data, f)



import dataclasses
import os
from pathlib import Path
import sys
from typing import TYPE_CHECKING, Optional, Union

import dacite
import dc_schema  # type: ignore

from bygg.output.output import TerminalStyle as TS
from bygg.output.output import output_plain

PYTHON_INPUTFILE = "Byggfile.py"
TOML_INPUTFILE = "Byggfile.toml"
YAML_INPUTFILE = "Byggfile.yml"

BYGGFILE_SUFFIXES = (".toml", ".yml", ".py")

DEFAULT_ENVIRONMENT_NAME = "_BYGG_DEFAULT_NULL_ENVIRONMENT"


@dataclasses.dataclass
class Settings:
    __doc__ = """
    The global settings object for Bygg.
    """

    def merge(self, other: "Settings"):
        """
        Merges the settings with another Settings object.
        """
        if other.default_action is not None:
            self.default_action = other.default_action
        if other.verbose is not None:
            self.verbose = other.verbose

    default_action: Optional[str] = None
    verbose: Optional[bool] = None


@dataclasses.dataclass
class ActionItem:
    """
    This is a representation of the Action class used for deserialising from YAML.
    The name of the action is the key in the dictionary in the config file.

    Parameters
    ----------
    description : str, optional
        A description of the action. Used in e.g. action listings. Default is None.
    message : str, optional
        A message to print when the action is executed. Default is None.
    inputs : list of str, optional
        A list of files that are used as input to the action. Default is None.
    outputs : list of str, optional
        A list of files that are generated by the action. Default is None.
    dependencies : list of str, optional
        A list of actions that must be executed before this action. Default is None.
    is_entrypoint : bool, optional
        Whether this action is an entrypoint. Entrypoints are actions that can be
        executed directly from the command line. If not set, this is treated as true by
        default in Byggfile.yml and false when used from Python. Default is None.
    environment : str, optional
        The environment for the action. Default is to run in the ambient environment.
    shell : str, optional
        The shell command to execute when the action is run. Default is None.
    """

    description: Optional[str] = None
    message: Optional[str] = None
    inputs: Optional[list[str]] = None
    outputs: Optional[list[str]] = None
    dependencies: Optional[list[str]] = None
    is_entrypoint: Optional[bool] = None
    environment: Optional[str] = DEFAULT_ENVIRONMENT_NAME
    shell: Optional[str] = None


@dataclasses.dataclass
class Environment:
    """
    A class used to represent a virtual environment that actions can be run in.

    Attributes
    ----------
    inputs : list[str]
        A list of files that are used as input to the environment. Typically pip
        requirements files, but can be any files.
    venv_directory : str
        The directory where the virtual environment is located. Will be recreated by
        Bygg if any of the inputs are modified.
    shell : str
        The shell command for creating the environment.
    byggfile : str
        The Python Byggfile that uses this environment. This is the entrypoint for where
        actions declared in Python are looked up. Optional.
    name : str, optional
        A human-friendly name for the environment. Used in e.g. help messages, by
        default None
    """

    inputs: list[str]
    venv_directory: str
    shell: str
    byggfile: Optional[str] = None
    name: Optional[str] = None


@dataclasses.dataclass
class Byggfile:
    class SchemaConfig:
        annotation = dc_schema.SchemaAnnotation(
            title="Schema for the configuration files for Bygg",
        )

    # Have to use Union here since dc_schema doesn't support the | notation
    actions: dict[str, Union[ActionItem, str]] = dataclasses.field(default_factory=dict)
    settings: Settings = dataclasses.field(default_factory=Settings)
    environments: dict[str, Environment] = dataclasses.field(default_factory=dict)


def has_byggfile() -> bool:
    """
    Checks if a Byggfile (in any format, including Python) exists in the current
    directory.

    Returns
    -------
    bool
        True if a Byggfile exists, False otherwise.
    """
    return len(get_config_files()) > 0 or os.path.isfile(PYTHON_INPUTFILE)


def get_config_files() -> list[Path]:
    """
    Checks if a static config file (a Byggfile that is not written in Python) is present
    in the current directory.
    """
    return [p for p in [Path(TOML_INPUTFILE), Path(YAML_INPUTFILE)] if p.is_file()]


def read_config_files() -> Byggfile:
    """Read all static config files and return them merged into one Byggfile object."""
    config_files = get_config_files()
    if not config_files:
        return Byggfile(actions={}, settings=Settings(), environments={})

    def action_item_hook(val):
        if isinstance(val, str):
            # Convert shortform actions to objects
            return ActionItem(shell=val)
        return dacite.from_dict(data=val, data_class=ActionItem)

    dacite_config = dacite.Config(
        strict=True,
        type_hooks={ActionItem: action_item_hook},
    )

    try:
        byggfile_objects: list[Byggfile] = []
        for cf in config_files:
            match cf.suffix:
                case ".toml":
                    import tomllib

                    with cf.open("rb") as toml_file:
                        byggfile_objects.append(
                            dacite.from_dict(
                                config=dacite_config,
                                data=tomllib.load(toml_file),
                                data_class=Byggfile,
                            )
                        )
                case ".yml":
                    import yaml

                    with cf.open("r", encoding="utf-8") as yaml_file:
                        byggfile_objects.append(
                            dacite.from_dict(
                                config=dacite_config,
                                data=yaml.safe_load(yaml_file),
                                data_class=Byggfile,
                            )
                        )
                case _:
                    raise ValueError(f"Unknown file extension {cf.suffix}")

        for bfo in byggfile_objects:
            for _, action in bfo.actions.items():
                if TYPE_CHECKING:
                    assert not isinstance(action, str)
                # Actions in byggfiles are entrypoints by default, unlike the actions
                # declared in Python files
                if action.is_entrypoint is None:
                    action.is_entrypoint = True
        return merge_byggfiles(byggfile_objects)

    except Exception as e:
        output_plain(
            TS.Fg.RED
            + " Error while reading configuration file."
            + TS.Fg.RESET
            + f" {e}"
        )
        sys.exit(1)


def merge_byggfiles(byggfiles: list[Byggfile]) -> Byggfile:
    """
    Merges a list of Byggfile objects into a single one.
    """
    merged_byggfile = Byggfile()
    for bf in byggfiles:
        merged_byggfile.actions.update(bf.actions)
        merged_byggfile.settings.merge(bf.settings)
        merged_byggfile.environments.update(bf.environments)
    return merged_byggfile


def dump_schema():
    import json
    import textwrap

    schema = dc_schema.get_schema(Byggfile)

    # Additional properties are not allowed, but dc_schema does not yet support this.
    # See https://github.com/Peter554/dc_schema/issues/6 .
    schema["additionalProperties"] = False

    classes_to_properties = {
        "ActionItem": "actions",
        "Environment": "environments",
        "Settings": "settings",
    }

    for k in schema["$defs"].keys():
        if k in classes_to_properties:
            schema["$defs"][k]["title"] = k
            schema["$defs"][k]["additionalProperties"] = False
            docstring = globals()[k].__doc__
            if docstring is not None:
                # Add the docstrings to the field types
                formatted_docstring = textwrap.dedent(docstring).strip()
                schema["$defs"][k]["description"] = formatted_docstring

                # Also add the docstrings to the containers
                schema["properties"][classes_to_properties[k]]["title"] = k
                schema["properties"][classes_to_properties[k]]["description"] = (
                    formatted_docstring
                )

    print(json.dumps(schema, indent=2))



import os
from pathlib import Path

import more_things.MoreActions  # noqa: F401 (imported for side effects)

from bygg.core.action import Action, ActionContext, action
from bygg.core.common_types import CommandStatus


@action(
    "testfile 1",
    inputs=["input1.txt"],
    outputs=["output1.txt"],
)
def transform_uppercase(ctx: ActionContext):
    print(f"Transforming: {ctx.inputs} -> {ctx.outputs}")
    if not ctx.inputs or not ctx.outputs:
        return CommandStatus(1, "No inputs or outputs.", None)

    for input_file in ctx.inputs:
        with open(input_file) as f:
            lines = f.readlines()
        p = Path(input_file)
        output_file = p.parent / p.name.replace("input", "output")

        with open(output_file, "w") as f:
            for line in lines:
                f.write(line.upper())

        return CommandStatus(0, "Transformed file.", None)

    return CommandStatus(1, "Something went wrong.", None)


@action(
    "dynamic",
    dynamic_dependency=lambda: f"{os.environ.get('DEBUG', 'False')}",
)
def dynamic_test(ctx: ActionContext):
    return CommandStatus(0, "Ran dynamic dependency action.", None)


Action(
    "transform",
    dependencies=["testfile 1", "testfile 2", "foo", "dynamic"],
    is_entrypoint=True,
)



import os
from pathlib import Path
import shutil
from tempfile import mkstemp

import pytest

from bygg.core.action import Action
from bygg.core.scheduler import Scheduler


def get_closed_tmpfile() -> Path:
    fd, path = mkstemp()
    os.close(fd)
    return Path(path)


@pytest.fixture
def scheduler_fixture():
    """Return a scheduler and a cache file."""
    scheduler = Scheduler()
    cache_file = get_closed_tmpfile()
    scheduler.init_cache(cache_file)
    Action._current_environment = "scheduler_fixture"
    yield (scheduler, cache_file)
    Action._current_environment = None
    scheduler.shutdown()


@pytest.fixture
def scheduler_single_action(scheduler_fixture):
    Action(
        name="action1",
        is_entrypoint=True,
    )
    return scheduler_fixture


@pytest.fixture
def scheduler_four_nonbranching_actions(scheduler_fixture):
    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(
        name="action2",
        dependencies=["action3"],
    )
    Action(
        name="action3",
        dependencies=["action4"],
    )
    Action(
        name="action4",
        dependencies=[],
    )
    return scheduler_fixture


@pytest.fixture
def scheduler_branching_actions(scheduler_fixture):
    Action(
        name="action1",
        dependencies=["action2", "action3"],
        is_entrypoint=True,
    )
    Action(
        name="action2",
        dependencies=["action4"],
    )
    Action(
        name="action3",
        dependencies=["action4"],
    )
    Action(
        name="action4",
        dependencies=[],
    )
    return scheduler_fixture


# .git is needed for hatch to be able to extract version when doing pip install
# from the sub environments
clean_bygg_tree_exclusions = (
    "__pycache__",
    ".bygg",
    ".jj",
    ".mypy_cache",
    ".nox",
    ".pytest_cache",
    ".ruff_cache",
    ".venv*",
    "foo",
    "bar",
    "t",
)


@pytest.fixture
def clean_bygg_tree(tmp_path):
    clean_path = tmp_path / "bygg"
    shutil.copytree(
        Path("."),
        clean_path,
        ignore=shutil.ignore_patterns(*clean_bygg_tree_exclusions),
    )
    return clean_path


@pytest.fixture(autouse=True)
def set_is_CI():
    os.environ["CI"] = "1"



from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Iterable, Literal, Optional, Self

from bygg.core.common_types import CommandStatus
from bygg.logutils import logger

if TYPE_CHECKING:
    from bygg.core.scheduler import Scheduler

SchedulingType = Literal["in-process", "processpool"]

# Function that returns a string that is included in the dependency digest. Returning
# None causes the value to be ignored.
DynamicDependency = Callable[[], str | None]


@dataclass
class ActionContext:
    name: str
    message: str | None
    inputs: set[str]
    outputs: set[str]
    dependencies: set[str]
    dynamic_dependency: Optional[DynamicDependency]
    is_entrypoint: bool
    scheduling_type: SchedulingType


Command = Callable[[ActionContext], CommandStatus]


class Action(ActionContext):
    """
    An action in the build graph.

    Parameters
    ----------
    name : str
        The name of the action.
    message : str, optional
        A message to display when the action is run. Default is None.
    inputs : Iterable[str], optional
        An iterable of paths that are inputs to the action. Default is None.
    outputs : Iterable[str], optional
        An iterable of paths that are outputs of the action. Default is None.
    dependencies : Iterable[str], optional
        An iterable of action names that this action depends on. Default is None.
    dynamic_dependency : DynamicDependency, optional
        A dynamic dependency of the action. Default is None.
    is_entrypoint : bool, optional
        Whether this action is an entrypoint to the build graph. Default is False.
    command : Command, optional
        Function to run. Default is None.
    scheduling_type : SchedulingType, optional
        The scheduling type for the action. Default is "processpool". Use "in-process"
        for small Python functions that finish quickly so that they can be run in the
        main process.
    description : str, optional
        A description of the action. Default is the docstring of the command if
        provided, else None.
    """

    # Set when Scheduler is initialised. Quotes around the type because of circular
    # import shenanigans.
    scheduler: "Scheduler | None" = None
    _current_environment: str | None = None

    command: Command | None
    dependency_files: set[str]

    def __init__(
        self,
        name: str,
        message: str | None = None,
        inputs: Optional[Iterable[str]] = None,
        outputs: Optional[Iterable[str]] = None,
        dependencies: Optional[Iterable[str | Self]] = None,
        dynamic_dependency: Optional[DynamicDependency] = None,
        is_entrypoint: bool = False,
        command: Command | None = None,
        scheduling_type: SchedulingType = "processpool",
        description: str | None = None,
        environment: Optional[str] = None,
    ):
        self.name = name
        self.message = message
        self.inputs = {*inputs} if inputs else set()
        self.outputs = {*outputs} if outputs else set()
        self.dependencies = {
            d.name if isinstance(d, Action) else d for d in (dependencies or [])
        }
        self.dynamic_dependency = dynamic_dependency
        self.is_entrypoint = is_entrypoint
        self.command = command
        self.scheduling_type = scheduling_type

        self.description = (
            description
            if description is not None
            else command.__doc__
            if command is not None and command.__doc__ is not None
            else None
        )

        self.environment = environment or Action._current_environment
        assert self.environment
        logger.info("Constructing Action for environment %s", self.environment)

        self.dependency_files = set()

        assert self.scheduler
        self.scheduler.build_actions[name] = self

    def __repr__(self):
        return f"""Action(name={self.name}, inputs=({
            ", ".join([str(i) for i in self.inputs])
        }), outputs=({", ".join([str(o) for o in self.outputs])}), dependencies=({
            ", ".join(self.dependencies)
        }) is_entrypoint={self.is_entrypoint})"""

    def __str__(self):
        return self.__repr__()


def action(
    name: str,
    *,
    message: Optional[str] = None,
    inputs: Optional[Iterable[str]] = None,
    outputs: Optional[Iterable[str]] = None,
    dependencies: Optional[Iterable[str | Action]] = None,
    dynamic_dependency: Optional[DynamicDependency] = None,
    scheduling_type: SchedulingType = "processpool",
    is_entrypoint: bool = False,
):
    """Decorator to define a Bygg action.

    Wraps the decorated function in an Action instance.

    Parameters
    ----------
    name : str
        The name of the action
    message : str, optional
        A message to display when the action is run, by default None
    inputs : Iterable[str], optional
        An iterable of input files, by default None
    outputs : Iterable[str], optional
        An iterable of output files, by default None
    dependencies : Iterable[str], optional
        An iterable of dependency actions, by default None
    dynamic_dependency : DynamicDependency, optional
        A dynamic dependency, by default None
    is_entrypoint : bool, optional
        Whether the action is an entrypoint, by default False
    scheduling_type : SchedulingType, optional
        The scheduling type for the action. Default is "processpool". Use "in-process"
        for small Python functions that finish quickly so that they can be run in the
        main process.
    description : str, optional
        A description of the action, by default None

    Returns
    -------
    Callable[[Callable], Action]
        A decorator that converts the decorated function into an Action
    """

    def create_action(func: Callable):
        return Action(
            name,
            message=message,
            inputs=inputs,
            outputs=outputs,
            dependencies=dependencies,
            dynamic_dependency=dynamic_dependency,
            is_entrypoint=is_entrypoint,
            scheduling_type=scheduling_type,
            command=func,
        )

    return create_action


def action_set(
    base_name: str,
    *,
    message: Optional[str] = None,
    file_pairs: Iterable[tuple[str, str]],
    extra_inputs: Optional[Iterable[str]] = None,
    dependencies: Optional[Iterable[str | Action]] = None,
    is_entrypoint: bool = False,
    scheduling_type: SchedulingType = "processpool",
    description: str | None = None,
):
    """
    Decorator for creating individual Actions from a list of input-output file pairs and
    a function. Creates one Action that depends on all of the individual Actions. It has
    name = base_name, which in turn can be used as a dependency for other Actions.

    Parameters
    ----------
    base_name : str
        Base name to use for generated actions
    message : str, optional
        Message for actions, by default None
    file_pairs : Iterable[tuple[str, str]]
        Input and output file pairs
    extra_inputs : Iterable[str], optional
        Extra input files, by default None
    dependencies : Iterable[str], optional
        Dependencies for actions, by default None
    is_entrypoint : bool, optional
        Whether the top-most action should be an entrypoint, by default False
    scheduling_type : SchedulingType, optional
        The scheduling type for the action. Default is "processpool". Use "in-process"
        for small Python functions that finish quickly so that they can be run in the
        main process.
    description : str, optional
        A description of the action, by default None

    Returns
    -------
    Callable[[Callable], None]
        A decorator that registers actions constructed around the decorated function
    """

    def create_actions(func: Callable):
        action_list = []
        for input_file, output_file in file_pairs:
            action_name = f"{base_name}::{output_file}"
            action_list.append(action_name)
            Action(
                action_name,
                message,
                inputs=[input_file] + (list(extra_inputs) if extra_inputs else []),
                outputs=[output_file],
                dependencies=dependencies,
                scheduling_type=scheduling_type,
                command=func,
            )
        Action(
            base_name,
            f"Action set {base_name}",
            dependencies=action_list,
            is_entrypoint=is_entrypoint,
            scheduling_type=scheduling_type,
            description=description,
        )

    return create_actions



from pathlib import Path
from typing import Literal

from bygg.core.action import Action
from bygg.core.cache import Cache
from bygg.core.dag import Dag, create_dag
from bygg.core.digest import calculate_dependency_digest, calculate_digest
from bygg.core.job import Job
from bygg.logutils import logger
from bygg.output.status_display import on_check_failed


class Scheduler:
    cache: Cache
    build_actions: dict[str, Action]

    job_graph: Dag
    ready_jobs: set[str]
    running_jobs: dict[str, Job]
    finished_jobs: dict[str, Job]

    started: bool
    always_make: bool
    check_inputs_outputs_set: set[str] | None

    def __init__(self):
        Action.scheduler = self
        self.cache = Cache()
        self.build_actions = {}
        self.job_graph = create_dag()
        self.ready_jobs = set()
        self.running_jobs = {}
        self.finished_jobs = {}
        self.started = False
        self.always_make = False
        self.check_inputs_outputs_set = None

    def init_cache(self, cache_file: Path):
        self.cache = Cache(cache_file)

    def prepare_run(self, entrypoint: str, check=False):
        self.job_graph.clear()
        self.ready_jobs = set()
        self.running_jobs = {}
        self.finished_jobs = {}

        self.job_graph.build_action_graph(
            self.build_actions, self.build_actions[entrypoint]
        )

        if check:
            # Do this check before dependency files are filled in from dependencies
            files_with_several_actions = {
                f: a
                for f, a in self.create_outputs_to_action_dict().items()
                if len(a) > 1
            }
            for f, a in files_with_several_actions.items():
                scapegoat = list(a)[0]
                on_check_failed(
                    "same_output_files",
                    self.build_actions[scapegoat],
                    f"The file {f} is in the output list from multiple actions: {', '.join(a)}",
                    "error",
                )

        # Fill the actions' dependency_files from self and their dependencies
        for action in self.build_actions.values():
            action.dependency_files.update(action.inputs)
            for dependency in action.dependencies:
                action.dependency_files.update(self.build_actions[dependency].outputs)
            # print(f"Action {action.name} inputs: {action._dependency_files}")

    def start_run(
        self, entrypoint: str, *, always_make: bool = False, check: bool = False
    ):
        self.always_make = always_make
        self.check_inputs_outputs_set = set() if check else None
        self.prepare_run(entrypoint, check)
        self.cache.load()
        self.started = True

    def create_outputs_to_action_dict(self):
        """Create a dictionary of outputs to actions"""
        res: dict[str, set[str]] = {}
        for job in self.job_graph.get_all_jobs():
            action = self.build_actions[job]
            for output in action.outputs:
                if output not in res:
                    res[output] = set()
                res[output].add(action.name)
        return res

    def shutdown(self):
        self.cache.save()

    def run_status(self) -> Literal["not started", "running", "finished", "failed"]:
        """Check the status of the run"""
        if not self.started:
            return "not started"
        if (
            len(
                [
                    x
                    for x in self.finished_jobs.values()
                    if x.status is not None and x.status.rc != 0
                ]
            )
            > 0
        ):
            return "failed"
        if len(self.job_graph) == 0:
            return "finished"
        return "running"

    def check_dirty(self, job_name: str) -> bool:
        """Check if a job needs to be built."""
        if self.always_make:
            return True

        action = self.build_actions[job_name]
        cached_digests = self.cache.get_digests(job_name)

        if (
            len(action.inputs) == 0
            and len(action.outputs) == 0
            and not action.dynamic_dependency
        ):
            # An action with neither inputs nor outputs will always be built
            logger.debug("Job '%s' is dirty (no inputs or outputs)", job_name)
            return True

        if (
            not cached_digests
            or not cached_digests.outputs_digest
            or not cached_digests.inputs_digest
        ):
            # No previous result, so we need to build
            logger.debug("Job '%s' is dirty (no previous result)", job_name)
            return True

        outputs_digest, files_were_missing = calculate_dependency_digest(action.outputs)
        if files_were_missing or cached_digests.outputs_digest != outputs_digest:
            # The output has changed, so we need to rebuild
            if files_were_missing:
                logger.debug("Job '%s' is dirty (output had missing files)", job_name)
            else:
                logger.debug("Job '%s' is dirty (output digest changed)", job_name)
            return True

        inputs_digest, files_were_missing = calculate_dependency_digest(
            action.dependency_files
        )

        # Calculate dynamic dependencies
        if action.dynamic_dependency:
            if not cached_digests.dynamic_digest:
                return True

            dd_result = action.dynamic_dependency()
            if (
                dd_result is None
                or calculate_digest([dd_result]) != cached_digests.dynamic_digest
            ):
                logger.debug("Job '%s' is dirty (dynamic dependency changed)", job_name)
                return True

        # TODO handle files_were_missing here? abort?
        if cached_digests.inputs_digest != inputs_digest:
            # The inputs have changed, so we need to rebuild
            logger.debug("Job '%s' is dirty (inputs changed)", job_name)
            return True

        logger.debug("Job '%s' is clean", job_name)
        return False

    def get_ready_jobs(self, batch_size: int = 0) -> list[Job]:
        """
        Create a batch of jobs and put them in the running pool. Returns all ready jobs
        if batch_size is 0.

        An empty job list may be returned even if there are more jobs left to run if all
        jobs in the batch were skipped. Job runners should continue polling for more
        jobs until the scheduler status is "finished".
        """
        if batch_size == 0 or len(self.ready_jobs) < batch_size:
            new_jobs = self.job_graph.get_ready_jobs(
                self.finished_jobs, self.running_jobs
            )
            dirty_jobs = []
            for job in new_jobs:
                if self.check_dirty(job):
                    dirty_jobs.append(job)
                else:
                    self.skip_job(job)
            self.ready_jobs.update(dirty_jobs)

        if len(self.ready_jobs) == 0:
            return []

        job_list: list[Job] = []
        for _ in range(
            len(self.ready_jobs)
            if batch_size == 0
            else min(batch_size, len(self.ready_jobs))
        ):
            action = self.build_actions[self.ready_jobs.pop()]
            new_job = Job(action)
            self.running_jobs[new_job.name] = new_job
            job_list.append(new_job)
            self.store_input_digests(new_job)

        if self.check_inputs_outputs_set is not None:
            # Check that no job has outputs that are inputs to an earlier job
            self.check_inputs_outputs_set.update(
                *[job.action.inputs for job in job_list],
                *[job.action.dependency_files for job in job_list],
            )
            for job_to_check in job_list:
                overlap = job_to_check.action.outputs & self.check_inputs_outputs_set
                if overlap:
                    plural = len(overlap) > 1
                    on_check_failed(
                        "check_inputs_outputs",
                        job_to_check.action,
                        f"Output{'s' if plural else ''} was declared as input{'s' if plural else ''} to earlier action: {', '.join(overlap)}",
                        "error",
                    )

        return job_list

    def skip_job(self, job_name: str):
        """Skip a job, remove it from the graph and don't send it to the runner."""
        self.job_graph.remove_node(job_name)

    def job_finished(self, job: Job):
        """Move a job from the running pool to the finished pool"""
        self.running_jobs.pop(job.name)
        self.finished_jobs[job.name] = job

        if job.status and job.status.rc == 0:
            self.job_graph.remove_node(job.name)
            self.store_output_digest(job)
        else:
            self.cache.remove_digests(job.name)

    def store_input_digests(self, job: Job):
        inputs_digest, _ = calculate_dependency_digest(job.action.dependency_files)
        dynamic_digest = (
            calculate_digest([dd])
            if job.action.dynamic_dependency and (dd := job.action.dynamic_dependency())
            else None
        )
        self.cache.set_digests(job.name, inputs_digest, "", dynamic_digest)

    def store_output_digest(self, job: Job):
        stored_digest = self.cache.get_digests(job.name)
        assert stored_digest is not None, (
            "At this point, the digest should already be in the cache since the job was started, or it's a bug"
        )

        outputs_digest, _ = calculate_dependency_digest(job.action.outputs)
        self.cache.set_digests(
            job.name,
            stored_digest.inputs_digest,
            outputs_digest,
            stored_digest.dynamic_digest,
        )



from bygg.cmd.argument_parsing import MaintenanceCommand
from bygg.cmd.configuration import Byggfile
from bygg.cmd.environments import remove_environments
from bygg.core.cache import Cache
from bygg.logutils import logger
from bygg.output.output import output_info


def perform_maintenance(configuration: Byggfile, commands: list[MaintenanceCommand]):
    logger.debug("Maintenance commands: %s", commands)
    # Sorting for testability and tidyness
    unique_cmds = sorted(set(commands))

    while unique_cmds and (cmd := unique_cmds.pop()):
        match cmd:
            case "remove_cache":
                output_info("Removing cache")
                Cache.reset()
            case "remove_environments":
                output_info("Removing environments")
                remove_environments(configuration)
            case _:
                raise ValueError(f"Unknown maintenance command '{cmd}'")



import pytest

from bygg.core.action import Action, ActionContext, CommandStatus
from bygg.core.scheduler import Scheduler


@pytest.fixture
def init_scheduler():
    Scheduler()
    Action._current_environment = "test_Action"
    yield
    Action._current_environment = None
    Action.scheduler = None


def test_Action_setup_empty(init_scheduler):
    action = Action(name="empty Action", inputs=None, outputs=None, dependencies=None)
    assert action.name == "empty Action"
    assert action.inputs == set()
    assert action.outputs == set()
    assert action.dependencies == set()
    assert action.is_entrypoint is False
    assert action.command is None


def test_Action_setup_with_dependencies(init_scheduler):
    action = Action(
        name="test Action",
        inputs=["input1", "input2"],
        outputs=["output1", "output2"],
        dependencies=["dependency1", "dependency2"],
        is_entrypoint=True,
    )
    assert action.name == "test Action"
    assert action.inputs == {"input1", "input2"}
    assert action.outputs == {"output1", "output2"}
    assert action.dependencies == {"dependency1", "dependency2"}
    assert action.is_entrypoint is True
    assert action.command is None


def test_Action_with_python_cmd(init_scheduler):
    def test_command(ctx: ActionContext):
        if not ctx.inputs:
            return CommandStatus(1, "No inputs.", None)

        return CommandStatus(0, "Executed successfully", None)

    action = Action(
        name="test with Python command",
        command=test_command,
        inputs=["input1", "input2"],
        outputs=["changed_file1", "changed_file2"],
    )

    assert action.command is not None
    assert action.command is test_command
    assert action.command(action).rc == 0
    assert action.command(action).message == "Executed successfully"



from pathlib import Path

import pytest

from bygg.output.job_output import HighlightConfig, highlight_log

short_gcc_log = """
main.c: In function ‘main’:
main.c:5:5: warning: implicit declaration of function ‘foo’ [-Wimplicit-function-declaration]
    foo();
    ^~~
main.c:5:5: warning: unused variable ‘a’ [-Wunused-variable]
    int a = 5;
    ^~~
main.c:6:5: warning: unused variable ‘b’ [-Wunused-variable]
    int b = 10;
    ^~~
main.c:7:5: warning: unused variable ‘c’ [-Wunused-variable]
    int c = a + b;
    ^~~
main.c:9:5: error: expected declaration or statement at end of input
    return 0;
    ^~~~~~
"""


def test_no_config():
    assert highlight_log(short_gcc_log) == short_gcc_log


config = [
    HighlightConfig(
        pattern=r"(warning:)",
        start="[START_WARNING]",
        end="[END_WARNING]",
    ),
    HighlightConfig(
        pattern=r"(error:)",
        start="[START_ERROR]",
        end="[END_ERROR]",
    ),
]


def test_highlight_warning(snapshot):
    print(highlight_log(short_gcc_log, config))
    assert highlight_log(short_gcc_log, config) == snapshot


testdata_path = Path("examples/failing_jobs/testdata/")
logfiles = ["gcc.log", "go.log", "make.log", "npm.log"]


@pytest.mark.parametrize("logfile", logfiles, ids=lambda x: x)
def test_highlight_warning_logfiles(snapshot, logfile):
    with open(testdata_path / logfile) as f:
        log = f.read()
    assert highlight_log(log, config) == snapshot



import pytest

from bygg.core.action import Action
from bygg.util import create_shell_command, filenames_from_pattern

test_cases_filenames_from_pattern = [
    (
        (
            "%.j2",
            "out_%",
            [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")],
        )
    ),
    (
        (
            "%.txt.j2",
            "out_%.txt",
            [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")],
        )
    ),
    (
        (
            "%.c",
            "%.o",
            [("foo.c", "foo.o"), ("bar.c", "bar.o")],
        )
    ),
    (
        (
            "%",
            "out_%",
            [("foo.txt", "out_foo.txt"), ("bar.txt", "out_bar.txt")],
        )
    ),
    (
        (
            "site/%.html.j2",
            "output/%.html",
            [
                ("site/foo.html.j2", "output/foo.html"),
                ("site/bar.html.j2", "output/bar.html"),
            ],
        )
    ),
    (
        "%.c",
        "%.o",
        [
            ("main.c", "main.o"),
            ("src/util.c", "src/util.o"),
            ("lib/helpers/string.c", "lib/helpers/string.o"),
        ],
    ),
    (
        "assets/%.scss",
        "css/%.css",
        [
            ("assets/main.scss", "css/main.css"),
            ("assets/components/button.scss", "css/components/button.css"),
        ],
    ),
    (
        "src/%.md",
        "build/docs/%.html",
        [
            ("src/readme.md", "build/docs/readme.html"),
            (
                "src/tutorials/getting-started.md",
                "build/docs/tutorials/getting-started.html",
            ),
        ],
    ),
    ("e%t", "c%r", [("eat", "car"), ("src/eat", "src/car"), ("east", "casr")]),
]


def construct_test_case_name(x: tuple):
    a, b, c = x
    return f"{a} | {b}"


@pytest.mark.parametrize(
    "test_case", test_cases_filenames_from_pattern, ids=construct_test_case_name
)
def test_filenames_from_pattern(test_case):
    in_pattern, out_pattern, input_output_files = test_case
    input_files = [x[0] for x in input_output_files]
    output_files = [x[1] for x in input_output_files]

    result = filenames_from_pattern(input_files, in_pattern, out_pattern)
    assert result.input_files == input_files
    assert result.output_files == output_files
    assert not result.unmatched_input_files


def test_shell_command():
    Action._current_environment = "test_shell_command"
    shell_command = create_shell_command("echo 'shell action output'")

    status = shell_command(Action("test context", inputs=set(), outputs=set()))
    assert status.rc == 0
    assert status.output == "shell action output\n"
    Action._current_environment = None



import os
from pathlib import Path
from tempfile import TemporaryDirectory, mkstemp

from bygg.core.cache import Cache, InputsOutputsDigests


def test_cache_load_non_existing():
    fd, path = mkstemp()
    os.close(fd)
    os.unlink(path)

    cache = Cache(Path(path))
    cache.load()
    assert cache.data is not None
    assert cache.data.digests == {}


def test_cache_load_empty():
    fd, path = mkstemp()
    cache = Cache(Path(path))
    cache.load()
    assert cache.data is not None
    assert cache.data.digests == {}
    os.close(fd)
    os.unlink(path)


def test_cache_write_and_load():
    with TemporaryDirectory() as d:
        cache = Cache(Path(d) / "cache.db")
        cache.load()
        assert cache.data is not None
        assert cache.data.digests == {}

        cache.data.digests["foo"] = InputsOutputsDigests("deadbeef", "f00", "f33d")
        cache.save()

        cache2 = Cache(Path(d) / "cache.db")
        cache2.load()
        assert cache2.data is not None
        assert cache2.data.digests == {
            "foo": InputsOutputsDigests("deadbeef", "f00", "f33d")
        }



import pytest

from bygg.cmd.argument_unparsing import unparse_args
from bygg.cmd.dispatcher import create_argument_parser

args = [
    (["-V"], ["--version"]),
    (["--version"], ["--version"]),
    (["--clean"], ["--clean"]),
    (["-l"], ["--list"]),
    (["--list"], ["--list"]),
    (["--tree"], ["--tree"]),
    (["-B"], ["--always-make"]),
    (["--always-make"], ["--always-make"]),
    (["-C", "foo/bar"], ["--directory=foo/bar"]),
    # more complex
    (["-C", "foo/bar", "-l"], ["--directory=foo/bar", "--list"]),
    (
        ["-C", "foo/bar", "action1", "-B"],
        ["--directory=foo/bar", "action1", "--always-make"],
    ),
]


@pytest.mark.parametrize("arg", args, ids=lambda x: " ".join(x[0]))
def test_unparse_args(arg):
    in_arg, out_arg = arg
    parser = create_argument_parser()
    parsed_args = parser.parse_args(in_arg)
    assert sorted(unparse_args(parser, parsed_args)) == sorted(out_arg)


def test_unparse_args_drop():
    parser = create_argument_parser()
    parsed_args = parser.parse_args(["-C", "foo/bar", "action1", "-B"])
    assert sorted(unparse_args(parser, parsed_args, drop=["directory"])) == sorted(
        ["action1", "--always-make"]
    )
    assert sorted(unparse_args(parser, parsed_args, drop=["always_make"])) == sorted(
        ["--directory=foo/bar", "action1"]
    )
    assert sorted(unparse_args(parser, parsed_args, drop=["actions"])) == sorted(
        ["--directory=foo/bar", "--always-make"]
    )
    assert sorted(
        unparse_args(parser, parsed_args, drop=["actions", "always_make", "directory"])
    ) == sorted([])



from dataclasses import dataclass
import shutil
from typing import Literal

from bygg.cmd.argument_parsing import ByggNamespace
from bygg.cmd.configuration import Byggfile
from bygg.core.action import Action
from bygg.core.common_types import JobStatus, Severity
from bygg.core.job import Job
from bygg.output.job_output import format_job_log
from bygg.output.output import (
    STATUS_TEXT_FIELD_WIDTH,
    output_error,
    output_info,
    output_warning,
    output_with_status_line,
)
from bygg.output.output import (
    TerminalStyle as TS,
)


def on_runner_status(message: str):
    output_info(message)


running_jobs: set[str] = set()


def format_queued_jobs_line(prefix: str) -> str:
    terminal_cols, _ = shutil.get_terminal_size()
    output = f"{prefix} {' '.join(running_jobs)}"
    if len(output) > terminal_cols:
        output = output[: terminal_cols - 3] + "..."
    return output


def format_result_status(status: JobStatus, field_width: int) -> str:
    match status:
        case "failed":
            return f"{TS.BOLD}{TS.Fg.GREEN}{'FAILED':>{field_width}}{TS.Fg.RESET}{TS.NOBOLD}"
        case "stopped":
            return f"{TS.BOLD}{TS.Fg.GREEN}{'STOPPED':>{field_width}}{TS.Fg.RESET}{TS.NOBOLD}"
        case "finished":
            return (
                f"{TS.BOLD}{TS.Fg.GREEN}{'OK':>{field_width}}{TS.Fg.RESET}{TS.NOBOLD}"
            )
        case _:
            return (
                f"{TS.BOLD}{TS.Fg.BLUE}{'OTHER':>{field_width}}{TS.Fg.RESET}{TS.NOBOLD}"
            )


max_name_length = 0


def get_on_job_status(args: ByggNamespace, configuration: Byggfile):
    def on_job_status(job_status: JobStatus, job: Job, jobs_count: tuple[int, int]):
        match job_status:
            case "skipped":
                pass
            case "running":
                running_jobs.add(job.name)
            case "failed" | "finished" | "stopped":
                running_jobs.discard(job.name)
                print_job_ended(job_status, job, jobs_count)
            case _:
                raise ValueError(f"Unhandled job status {job_status}")

    def print_job_ended(job_status: JobStatus, job: Job, jobs_count: tuple[int, int]):
        global max_name_length
        max_name_length = max(len(job.name), max_name_length)
        status_code_message = f"[{job.status.rc}] " if job.status else "?"
        status_message = job.status.message if job.status and job.status.message else ""
        message_part = f"{status_code_message if job.status and job.status.rc else ''}{status_message}"

        total_job_count_length = len(str(jobs_count[1]))
        job_count_info = f" ({jobs_count[0]:>{total_job_count_length}}/{jobs_count[1]})"

        if args.verbose or configuration.settings.verbose:
            result_status = format_result_status(job_status, 0)
            formatted_log = format_job_log(job)
            log = f"\n{formatted_log}" if formatted_log else ""
            output_with_status_line(
                format_queued_jobs_line(job_count_info),
                f"{result_status} {job.name}{(' : ' + message_part) if len(message_part) > 0 else ''}{log}",
            )
        else:
            result_status = format_result_status(job_status, STATUS_TEXT_FIELD_WIDTH)
            output_with_status_line(
                format_queued_jobs_line(job_count_info),
                f"{result_status} {job.name:<{max_name_length}}{(' : ' + message_part) if len(message_part) > 0 else ''}",
            )

    return on_job_status


CheckRule = Literal["check_inputs_outputs", "output_file_missing", "same_output_files"]
"""
The different rules that can be checked.

    check_inputs_outputs: Check that earlier actions don't need outputs from subsequent
    actions.

    output_file_missing: Check that actions create the files that they declare as
    outputs.

    same_output_files: Check that different actions don't produce the same output files.
"""


@dataclass
class CheckStatus:
    rule_name: CheckRule
    action: Action
    status_text: str
    severity: Severity


failed_checks: list[CheckStatus] = []


def on_check_failed(
    rule_name: CheckRule, action: Action, status_text: str, severity: Severity
):
    failed_checks.append(CheckStatus(rule_name, action, status_text, severity))


def output_check_results():
    status = True
    output_error("The following checks reported issues:")
    for c in failed_checks:
        compound_status = f"{c.rule_name} :: {c.action.name} :: {c.status_text}"
        match c.severity:
            case "error":
                status = False
                output_error(compound_status)
            case "warning":
                status = False
                output_warning(compound_status)
            case "info":
                continue
            case _:
                raise ValueError(f"Unhandled severity {c.severity}")

    return status



import argparse
import os
from pathlib import Path
import shutil
import sys
import textwrap
from typing import Any, Generator

from argcomplete.completers import BaseCompleter, DirectoriesCompleter
from argcomplete.finders import CompletionFinder

from bygg.cmd.configuration import BYGGFILE_SUFFIXES
from bygg.output.output import output_plain


class ByggfileDirectoriesCompleter(DirectoriesCompleter):
    """
    A completer for directories that contain Bygg files.
    """

    # TODO: error in typing in argcomplete, see https://github.com/rikardg/bygg/issues/192.
    def __call__(self, prefix, **kwargs) -> Generator[str, Any, None]:  # type: ignore
        directories = super().__call__(prefix, **kwargs)
        byggfile_dirs = set()
        for dir in directories:
            byggfile_dirs |= {
                str(b.parent)
                for b in Path(dir).rglob("Byggfile.*")
                if b.suffix in BYGGFILE_SUFFIXES
            }
        for f in sorted(list(byggfile_dirs)):
            yield f


def construct_already_there(
    parser: argparse.ArgumentParser | Any, options: set[str]
) -> set[str]:
    """
    Construct a set of options that are already present among the command line
    arguments. Adds both the short and long versions if they exist.
    """
    already_there_set: set[str] = set()
    for action in parser._get_optional_actions():
        option_strings = set(action.option_strings)
        if option_strings & options:
            already_there_set.update(option_strings)
    return already_there_set


singular_options = {
    "--completions",
    "--dump-schema",
    "--help",
    "--version",
    "-h",
    "-V",
}

directory_option = {"-C", "--directory"}


class ByggCompletionFinder(CompletionFinder):
    """
    Override _get_completions because we want custom completion patterns that are not
    expressed in argparse.
    """

    def _get_completions(
        self, comp_words, cword_prefix, cword_prequote, last_wordbreak_pos
    ) -> list[str]:
        comp_words_set = set(comp_words)

        # Stop completions if we already have a "singular" option, i.e. an option that
        # should not be followed by anything else.
        if comp_words_set & singular_options:
            return []

        # Get completions from argparse.
        completions = set(
            super()._get_completions(
                comp_words, cword_prefix, cword_prequote, last_wordbreak_pos
            )
        )

        # If the last word is a directory, the default completions will be from the
        # directories completer, so return these.
        if comp_words[-1] in directory_option:
            return list(completions)

        action_completions = list(filter(lambda x: not x.startswith("-"), completions))
        already_there = construct_already_there(self._parser, comp_words_set)

        if comp_words_set & {"--tree", "--clean"}:
            # Actions if we have them.
            if action_completions and not cword_prefix.startswith("-"):
                return action_completions
            return list(completions & directory_option - already_there)

        if comp_words_set & {"--list"}:
            # Only complete directories.
            return list(completions & directory_option - already_there)

        # Return action completions if they exist and user has not entered a '-'.
        if action_completions and not cword_prefix.startswith("-"):
            return action_completions

        return list(completions - already_there)


def do_completion(parser: argparse.ArgumentParser):
    if "--is_restarted_with_env" in sys.argv:
        return

    completer = ByggCompletionFinder()
    completer(parser)


def is_completing():
    return os.environ.get("_ARGCOMPLETE") is not None


def generate_shell_completions(print_and_exit: bool):
    bygg_bin = Path(sys.argv[0])
    bygg_completions = bygg_bin.resolve().parent / "_bygg_completions.sh"
    if not bygg_completions.exists():
        import argcomplete.shell_integration

        with open(bygg_completions, "w") as f:
            f.write(argcomplete.shell_integration.shellcode([str(bygg_bin.name)]))

    if print_and_exit:
        terminal_cols, _ = shutil.get_terminal_size()
        indent_string = "# "
        width = min(terminal_cols, 80) - len(indent_string)
        description = f"""\
            A shell completions script has been generated in {bygg_completions}. It will
            pick up the bygg that is in PATH. Add the following line to your shell's
            startup script to load completions:
        """
        output_plain(
            textwrap.indent(
                textwrap.fill(
                    textwrap.dedent(description),
                    width=width,
                    break_long_words=False,
                    break_on_hyphens=False,
                ),
                indent_string,
            )
        )
        output_plain(f"\nsource {bygg_completions}")

        sys.exit(0)


# TODO: error in typing in argcomplete, see https://github.com/rikardg/bygg/issues/192.
# pyright: reportIncompatibleMethodOverride=warning
class EntrypointCompleter(BaseCompleter):
    def __call__(
        self,
        *,
        prefix,
        action: argparse.Action,
        parser: argparse.ArgumentParser,
        parsed_args: argparse.Namespace,
        **kwargs,
    ):
        import textwrap

        from bygg.cmd.dispatcher import parent_dispatcher

        subprocess_data = parent_dispatcher(parser, parsed_args)[0]
        if not subprocess_data:
            return {}

        # There can technically only be one default_action, but easy way to consolidate:
        default_actions = {
            ipc_data.list.default_action
            for env, ipc_data in subprocess_data.items()
            if ipc_data.list is not None
        }

        env_actions = [
            ipc_data.list.actions
            for env, ipc_data in subprocess_data.items()
            if ipc_data.list is not None
        ]

        actions = {
            name: textwrap.fill(
                f"{'(default) ' if name in default_actions else ''}{description}", 7000
            )
            for action_set in env_actions
            for name, description in action_set.items()
            if name not in parsed_args.actions
        }

        return actions



import argparse
import dataclasses
from typing import Optional

import pytest

from bygg.cmd.argument_parsing import ByggNamespace, create_argument_parser
from bygg.cmd.configuration import Byggfile
from bygg.cmd.datastructures import ByggContext, SubProcessIpcData
from bygg.cmd.tree import print_tree, tree_collect_for_environment
from bygg.core.action import Action
from bygg.core.runner import ProcessRunner
from bygg.core.scheduler import Scheduler


@dataclasses.dataclass
class ArgparseFixtureData:
    parser: argparse.ArgumentParser
    args: argparse.Namespace
    bygg_namespace: ByggNamespace


@pytest.fixture
def create_byggcontext():
    def instantiate_context(scheduler: Scheduler, argv: Optional[list[str]] = None):
        parser = create_argument_parser()
        args = parser.parse_args(argv or [])
        return ByggContext(
            ProcessRunner(scheduler),
            scheduler,
            Byggfile(),
            parser,
            args,
            ByggNamespace(**vars(args)),
            SubProcessIpcData(),
        )

    yield instantiate_context


def test_tree_single_action(
    scheduler_single_action, create_byggcontext, capsys, snapshot
):
    scheduler, _ = scheduler_single_action
    ctx = create_byggcontext(scheduler)
    assert Action._current_environment
    print_tree(
        tree_collect_for_environment(ctx, Action._current_environment), ["action1"]
    )
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot


def test_tree_four_nonbranching_actions(
    scheduler_four_nonbranching_actions, create_byggcontext, capsys, snapshot
):
    scheduler, _ = scheduler_four_nonbranching_actions
    ctx = create_byggcontext(scheduler)
    assert Action._current_environment
    print_tree(
        tree_collect_for_environment(ctx, Action._current_environment), ["action1"]
    )
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot


def test_tree_branching_actions(
    scheduler_branching_actions, create_byggcontext, capsys, snapshot
):
    scheduler, _ = scheduler_branching_actions
    ctx = create_byggcontext(scheduler)
    assert Action._current_environment
    print_tree(
        tree_collect_for_environment(ctx, Action._current_environment), ["action1"]
    )
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot



from pathlib import Path
import time

from bygg.core.action import ActionContext, action
from bygg.core.common_types import CommandStatus


# Simulate e.g. fetching variables from a CI system. Doing it twice is not a practical
# use case, but works for testing.
@action(
    name="restart_once",
    is_entrypoint=True,
    outputs=["restart_status.txt"],
    dynamic_dependency=lambda: str(time.time()),
)
def restart_once(ctx: ActionContext):
    indicator_file = Path("restart_status.txt")
    if indicator_file.exists():
        with indicator_file.open("r", encoding="utf-8") as f:
            content = f.read()
        if content == "1":
            with indicator_file.open("w", encoding="utf-8") as f:
                content = f.write("2")
            return CommandStatus(0, "Restarting 1", None, "restart_build")
        if content == "2":
            return CommandStatus(0, "Proceeding, was up-to-date", None)
    else:
        with indicator_file.open("w", encoding="utf-8") as f:
            content = f.write("1")
    return CommandStatus(0, "Restarting", None, "restart_build")


@action(name="another_action", inputs=["has_run_once.txt"])
def another_action(ctx: ActionContext):
    return CommandStatus(0, "Finished", None)



import pytest

from bygg.core.action import Action, ActionContext, CommandStatus

pytestmark = pytest.mark.scheduler


def test_scheduler_single_action(scheduler_single_action):
    scheduler, _ = scheduler_single_action

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 1
    assert len(scheduler.job_graph) == 1

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"


def test_scheduler_nonbranching(scheduler_four_nonbranching_actions):
    scheduler, _ = scheduler_four_nonbranching_actions

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action4"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action3"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 2
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 4


def test_scheduler_branching(scheduler_branching_actions):
    scheduler, _ = scheduler_branching_actions

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    jobs = scheduler.get_ready_jobs(1)
    assert jobs
    job = jobs[0]
    assert job.action.name == "action4"
    assert not scheduler.get_ready_jobs(1)
    assert scheduler.run_status() == "running"

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3
    assert scheduler.run_status() == "running"

    job1 = scheduler.get_ready_jobs(1)[0]
    assert job1 is not None
    assert job1.action.name == "action2" or job1.action.name == "action3"
    job1.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    job2 = scheduler.get_ready_jobs(1)[0]
    assert job2 is not None
    assert job2.action.name == "action2" or job2.action.name == "action3"
    job2.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    assert job1.name != job2.name

    assert not scheduler.get_ready_jobs(1)
    scheduler.job_finished(job1)
    scheduler.job_finished(job2)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job is not None
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 4


def test_scheduler_branching_one_failed(scheduler_branching_actions):
    scheduler, _ = scheduler_branching_actions

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    job = scheduler.get_ready_jobs(1)[0]
    assert job is not None
    assert job.action.name == "action4"
    assert not scheduler.get_ready_jobs(1)
    assert scheduler.run_status() == "running"

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3

    job1 = scheduler.get_ready_jobs(1)[0]
    assert job1 is not None
    assert job1.action.name == "action2" or job1.action.name == "action3"
    job1.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    job2 = scheduler.get_ready_jobs(1)[0]
    assert job2 is not None
    assert job2.action.name == "action2" or job2.action.name == "action3"
    job2.status = CommandStatus(1, "Failed", None)
    assert scheduler.run_status() == "running"

    assert job1.name != job2.name

    assert not scheduler.get_ready_jobs(1)
    scheduler.job_finished(job1)
    scheduler.job_finished(job2)
    assert scheduler.run_status() == "failed"

    assert not scheduler.get_ready_jobs(1)
    assert len(scheduler.ready_jobs) == 0
    assert len(scheduler.job_graph) == 2
    assert scheduler.run_status() == "failed"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 3


def test_scheduler_dynamic_dependency(scheduler_fixture):
    scheduler, _ = scheduler_fixture

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"
    scheduler.start_run("action1")
    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    job = scheduler.get_ready_jobs()[0]
    assert job
    assert job.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    # check again that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 1

    job = scheduler.get_ready_jobs()[0]
    assert job
    assert len(scheduler.job_graph) == 1
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0


def test_scheduler_dynamic_dependency_two_runs(scheduler_fixture):
    scheduler, cache_file = scheduler_fixture

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 2

    scheduler.shutdown()

    # Second run

    scheduler.__init__()
    scheduler.init_cache(cache_file)

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    jobs = scheduler.get_ready_jobs()
    assert len(jobs) == 0
    assert scheduler.run_status() == "running"

    # Ask again since the job should have been skipped and the scheduler is still
    # running:
    job = scheduler.get_ready_jobs()[0]
    assert job
    assert len(scheduler.job_graph) == 1

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0


def test_scheduler_single_file(scheduler_fixture, tmp_path):
    scheduler, cache_file = scheduler_fixture

    def action1(ctx: ActionContext):
        for output in ctx.outputs:
            with open(output, "w") as f:
                f.write("file content")
        return CommandStatus(0, "Executed successfully", None)

    Action(
        name="file",
        is_entrypoint=True,
        inputs=[],
        outputs=[str(tmp_path / "file1")],
        command=action1,
    )

    scheduler.start_run("file")

    job = scheduler.get_ready_jobs(1)[0]
    job.status = job.action.command(job.action)
    scheduler.job_finished(job)

    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    scheduler.shutdown()

    # Second run

    scheduler.__init__()
    scheduler.init_cache(cache_file)

    Action(
        name="file",
        is_entrypoint=True,
        inputs=[],
        outputs=[str(tmp_path / "file1")],
        command=action1,
    )

    scheduler.start_run("file")
    job_list = scheduler.get_ready_jobs(1)

    assert len(job_list) == 0
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"


def test_scheduler_single_file_changed(scheduler_fixture, tmp_path):
    scheduler, cache_file = scheduler_fixture
    infile = tmp_path / "file0"
    outfile = tmp_path / "file1"

    def action1(ctx: ActionContext):
        for output in ctx.outputs:
            with open(output, "w") as f:
                f.write("file content")
        return CommandStatus(0, "Executed successfully", None)

    Action(
        name="file",
        is_entrypoint=True,
        inputs=[str(infile)],
        outputs=[str(outfile)],
        command=action1,
    )

    infile.write_text("infile content")

    scheduler.start_run("file")

    job = scheduler.get_ready_jobs(1)[0]
    job.status = job.action.command(job.action)
    scheduler.job_finished(job)

    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    scheduler.shutdown()

    # Second run

    scheduler.__init__()
    scheduler.init_cache(cache_file)

    Action(
        name="file",
        is_entrypoint=True,
        inputs=[str(infile)],
        outputs=[str(outfile)],
        command=action1,
    )

    scheduler.start_run("file")
    job_list = scheduler.get_ready_jobs(1)

    assert len(job_list) == 0
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    # Third run, modify input file

    infile.write_text("modified infile content")

    scheduler.__init__()
    scheduler.init_cache(cache_file)

    Action(
        name="file",
        is_entrypoint=True,
        inputs=[str(infile)],
        outputs=[str(outfile)],
        command=action1,
    )

    scheduler.start_run("file")
    job_list = scheduler.get_ready_jobs(1)

    assert len(job_list) == 1
    assert len(scheduler.job_graph) == 1

    job.status = job.action.command(job.action)
    scheduler.job_finished(job)

    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"



import argparse
from dataclasses import dataclass, field
import os
from typing import Optional

from bygg.cmd.argument_parsing import ByggNamespace
from bygg.cmd.configuration import Byggfile
from bygg.core.common_types import CommandStatus
from bygg.core.runner import ProcessRunner
from bygg.core.scheduler import Scheduler


@dataclass
class SubProcessIpcDataList:
    """Holds the data for --list from a subprocess."""

    actions: dict[str, str]
    default_action: Optional[str] = None


@dataclass
class SubProcessIpcDataTree:
    """Holds the data for --tree from a subprocess."""

    actions: dict[str, str]


@dataclass
class SubProcessIpcData:
    """Holds the results for metadata from a subprocess."""

    # Actions that were found
    found_actions: set[str] = field(default_factory=set)
    list: Optional[SubProcessIpcDataList] = None
    tree: Optional[SubProcessIpcDataTree] = None
    return_code: int = 0
    found_input_files: set[str] = field(default_factory=set)
    failed_jobs: dict[str, CommandStatus] = field(default_factory=dict)


@dataclass
class ByggContext:
    """Container for various state"""

    runner: ProcessRunner
    scheduler: Scheduler
    configuration: Byggfile
    parser: argparse.ArgumentParser
    args_namespace: argparse.Namespace
    bygg_namespace: ByggNamespace
    ipc_data: SubProcessIpcData


@dataclass
class EntryPoint:
    def __repr__(self):
        return self.name

    name: str
    description: str


NO_DESCRIPTION = "No description"


def get_entrypoints(ctx: ByggContext, environment_name: str) -> list[EntryPoint]:
    show_all = os.environ.get("BYGG_DEV")
    return [
        EntryPoint(x.name, x.description or NO_DESCRIPTION)
        for x in ctx.scheduler.build_actions.values()
        if (x.is_entrypoint or show_all) and x.environment == environment_name
    ]



from pathlib import Path
import shlex
import subprocess

import pytest

from bygg.system_helpers import change_dir


def dummy_exit_method(status):
    pass


def open_raise(*args, **kwargs):
    raise NotImplementedError


@pytest.fixture
def completion_tester(monkeypatch, tmp_path):
    def tester(testcase):
        prefix = "bygg"
        prefixed_testcase = f"{prefix} {testcase}"
        outfile = tmp_path / "completion.out"

        # Simulate being run from the completion script:
        monkeypatch.setenv("_ARGCOMPLETE", "1")
        monkeypatch.setenv("COMP_LINE", prefixed_testcase)
        monkeypatch.setenv("COMP_POINT", str(len(prefixed_testcase)))
        monkeypatch.setenv("_ARGCOMPLETE_STDOUT_FILENAME", str(outfile))

        subprocess.run(
            shlex.split(prefixed_testcase),
            capture_output=True,
            encoding="utf-8",
        )

        if not outfile.exists():
            return ""

        with open(outfile, "r", encoding="utf-8") as f:
            return f.read()

    return tester


args = [
    # Complete simple arguments:
    ("--vers", "--version "),
    ("--verb", "--verbose "),
    ("--cl", "--clean "),
    ("--h", "--help "),
    # No completions after these arguments:
    ("-h ", ""),
    ("--help ", ""),
    ("-V ", ""),
    ("--version ", ""),
    ("--dump-schema ", ""),
    ("--completions ", ""),
]


@pytest.mark.parametrize("arg", args, ids=lambda x: x[0])
def test_completions(completion_tester, arg):
    testcase, correct_result = arg
    testresult = completion_tester(testcase)
    assert correct_result == testresult


# Directories:
directories_completions = [
    "-C ",
    "--directory ",
]


@pytest.mark.parametrize("arg", directories_completions, ids=lambda x: x)
def test_directory_completions(completion_tester, arg, snapshot, monkeypatch):
    monkeypatch.chdir("examples")
    testcase = arg
    testresult = completion_tester(testcase)
    assert sorted(testresult.split("\x0b")) == snapshot


actions_completions = [p for p in Path("examples").iterdir() if p.is_dir()]


@pytest.mark.parametrize(
    "bygg_dev", [False, True], ids=["BYGG_DEV_unset", "BYGG_DEV_set"]
)
@pytest.mark.parametrize("example_dir", actions_completions, ids=lambda x: str(x.name))
@pytest.mark.parametrize(
    "testcase_template",
    ["-C {dir} ", "--directory {dir} ", ""],
    ids=["-C", "--directory", "cwd"],
)
def test_actions_completions(
    completion_tester,
    example_dir,
    snapshot,
    clean_bygg_tree,
    monkeypatch,
    bygg_dev,
    testcase_template,
):
    if bygg_dev:
        monkeypatch.setenv("BYGG_DEV", "1")

    # Empty template means no -C or --directory argument
    should_chdir = len(testcase_template) == 0
    testcase = (
        testcase_template
        if should_chdir
        else testcase_template.format(dir=str(example_dir))
    )

    with change_dir(
        (clean_bygg_tree / example_dir) if should_chdir else clean_bygg_tree
    ):
        testresult = completion_tester(testcase)
        assert sorted(testresult.split("\x0b")) == snapshot



import os
import shutil

from bygg.core.action import ActionContext, action_set
from bygg.core.common_types import CommandStatus

static_file_pairs = [
    ("files/a.txt", "files/out/a.txt"),
    ("files/b.txt", "files/out/b.txt"),
    ("files/c.txt", "files/out/c.txt"),
]


@action_set(
    "testfiles_static",
    message="Transforming file",
    file_pairs=static_file_pairs,
    is_entrypoint=True,
    description="Transforms a file to another file (testcase for action_set)",
)
def transform_testfile(ctx: ActionContext):
    if not ctx.inputs or not ctx.outputs:
        return CommandStatus(1, "No inputs or outputs.", None)
    for input_file, output_file in zip(ctx.inputs, ctx.outputs):
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        shutil.copy(input_file, output_file)
    return CommandStatus(0, f"{ctx.inputs} -> {ctx.outputs}", None)



from dataclasses import dataclass
from pathlib import Path
import subprocess

import pytest


@dataclass
class TestcaseParameters:
    __test__ = False
    name: str
    actions: list[str]


testcases_dir = Path("testcases")

testcases: list[TestcaseParameters] = [
    TestcaseParameters("action_set", ["testfiles_static"]),
    TestcaseParameters("restart_build", ["restart_once"]),
]


filetree_ignore = (
    "__pycache__",
    ".bygg",
)


def filetree(path: str):
    return sorted(
        [
            str(p.relative_to(path))
            for p in Path(path).rglob("**/*")
            if not any(sub in str(p) for sub in filetree_ignore)
        ]
    )


@pytest.mark.parametrize("testcase", testcases, ids=lambda x: x.name)
def test_list(snapshot, clean_bygg_tree, testcase):
    example_path = clean_bygg_tree / testcases_dir / testcase.name

    process = subprocess.run(
        ["bygg", "--list"],
        cwd=clean_bygg_tree / testcases_dir / testcase.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot
    assert filetree(example_path) == snapshot

    for action in testcase.actions:
        process = subprocess.run(
            ["bygg", action],
            cwd=clean_bygg_tree / testcases_dir / testcase.name,
            capture_output=True,
            encoding="utf-8",
        )
        assert process.returncode == 0
        assert filetree(example_path) == snapshot

        process = subprocess.run(
            ["bygg", action, "--clean"],
            cwd=clean_bygg_tree / testcases_dir / testcase.name,
            capture_output=True,
            encoding="utf-8",
        )
        assert process.returncode == 0
        assert filetree(example_path) == snapshot



import shutil
import textwrap

from bygg.cmd.configuration import DEFAULT_ENVIRONMENT_NAME
from bygg.cmd.datastructures import (
    ByggContext,
    EntryPoint,
    SubProcessIpcData,
    SubProcessIpcDataList,
    get_entrypoints,
)
from bygg.output.output import (
    TerminalStyle as TS,
)

list_actions_style = "B"

HEADER = f"{TS.BOLD}Available actions:{TS.RESET}"


def list_collect_for_environment(
    ctx: ByggContext, environment_name: str
) -> SubProcessIpcDataList:
    """Collects the currently loaded entrypoints"""
    entrypoints = get_entrypoints(ctx, environment_name)
    sorted_actions = sorted(entrypoints, key=lambda x: x.name)
    default_action_name = ctx.configuration.settings.default_action

    return SubProcessIpcDataList(
        actions={x.name: x.description for x in sorted_actions},
        default_action=default_action_name,
    )


def list_actions_A(actions: list[EntryPoint]) -> list[str]:
    terminal_cols, terminal_rows = shutil.get_terminal_size()

    output: list[str] = []
    output.append("")
    section_indent = 0
    separator = " : "
    max_name_width = max([len(x.name) for x in actions])
    width = min(terminal_cols, 80)
    subsequent_indent = " " * (section_indent + max_name_width + len(separator))
    for action in actions:
        description = f"{TS.BOLD}{action.name: <{max_name_width}}{TS.RESET}{separator}{action.description}"
        output.extend(
            textwrap.wrap(
                description,
                width=width,
                initial_indent=" " * section_indent,
                subsequent_indent=subsequent_indent,
            )
        )
        output.append("")
    return output


def list_actions_B(
    actions: list[EntryPoint], default_action_name: str | None
) -> list[str]:
    output: list[str] = []
    output.append("")
    for action in actions:
        default_action_suffix = (
            " (default)" if action.name == default_action_name else ""
        )
        output.append(f"{TS.BOLD}{action.name}{default_action_suffix}{TS.RESET}")
        output.append(
            textwrap.fill(
                action.description,
                initial_indent="    ",
                subsequent_indent="    ",
            )
        )
        output.append("")
    return output


def print_actions(ctx: ByggContext, environment_data: dict[str, SubProcessIpcData]):
    """Prints a list of actions from different environments."""
    display_environment_names = (
        len(
            [
                len(v.list.actions)
                for k, v in environment_data.items()
                if v.list and len(v.list.actions) > 0
            ]
        )
        > 1
    )

    output = [HEADER]

    if display_environment_names:
        output.append("")

    for env, data in sorted(
        environment_data.items(),
        key=lambda x: (0, x[0]) if x[0] == DEFAULT_ENVIRONMENT_NAME else (1, x[0]),
    ):
        if not data.list:
            continue

        if display_environment_names:
            # Get the human-friendly name or fall back to the environment key:
            environment = ctx.configuration.environments.get(env)
            if env == DEFAULT_ENVIRONMENT_NAME:
                environment_name = "No environment"
            elif environment and environment.name:
                environment_name = environment.name
            else:
                environment_name = env

            output.append(f"~~ {environment_name} ~~")

        output += list_actions_B(
            [EntryPoint(name, descr) for name, descr in data.list.actions.items()],
            data.list.default_action,
        )
    print("\n".join(output))



from dataclasses import dataclass, field
from pathlib import Path
import re
import subprocess
import sys

import pytest

from bygg.cmd.argument_parsing import fill_help_text
from bygg.core.cache import DEFAULT_DB_FILE


@pytest.mark.help
@pytest.mark.parametrize(
    "python_version", [f"{sys.version_info.major}.{sys.version_info.minor}"]
)
def test_help(snapshot, clean_bygg_tree, python_version):
    process = subprocess.run(
        ["bygg", "--help"],
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot


@dataclass
class HelpText:
    name: str
    text: str


help_texts = [
    HelpText(
        "basic",
        "This is a basic help text.",
    ),
    HelpText(
        "multiline",
        """
This is a multiline help text.

Line 2. Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.

Line 3. Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
""",
    ),
    HelpText(
        "lists",
        """
This is a multiline help text with lists.

- Line 2. Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
- Line 3. Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
""",
    ),
]


@pytest.mark.parametrize("help_text", help_texts, ids=lambda x: x.name)
def test_help_formatting(snapshot, help_text):
    assert fill_help_text(help_text.text, 80, "") == snapshot
    assert fill_help_text(help_text.text, 80, "  ") == snapshot


@dataclass
class ExampleParameters:
    name: str
    list_rc: int = 0
    tree_rc: int = 0
    build_rc: int = 0
    environments: list[str] = field(default_factory=list)
    build_yields_cache: bool = True
    actions_for_all_environments: list[str] = field(default_factory=list)


examples = [
    ExampleParameters("checks"),
    ExampleParameters(
        "environments",
        environments=[".venv", ".venv1", ".venv2"],
        actions_for_all_environments=["default_action", "action1", "action2"],
    ),
    ExampleParameters("failing_jobs", build_rc=1),
    ExampleParameters("only_python", tree_rc=1, build_rc=1, build_yields_cache=False),
    ExampleParameters("parametric"),
    ExampleParameters("taskrunner"),
    ExampleParameters("trivial"),
]

examples_dir = Path("examples")


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_list(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--list"],
        cwd=clean_bygg_tree / examples_dir / example.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.list_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_list_directory(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--list", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.list_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_tree(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--tree"],
        cwd=clean_bygg_tree / examples_dir / example.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.tree_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_tree_directory(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--tree", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.tree_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_tree_non_existing_action(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--tree", "no_such_action"],
        cwd=clean_bygg_tree / examples_dir / example.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 1
    assert process.stdout == snapshot


# TODO more tests:
# --tree ACTION (one and several existing )


def test_dump_schema(snapshot):
    process = subprocess.run(
        ["bygg", "--dump-schema"],
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot


def test_schema_dump_is_uptodate():
    process = subprocess.run(
        ["bygg", "--dump-schema"],
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0

    with open("schemas/Byggfile_yml_schema.json", "r", encoding="utf-8") as f:
        schema = f.read()

    assert process.stdout == schema


def test_version():
    process = subprocess.run(
        ["bygg", "--version"],
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout[8:14] == " bygg "
    # Version output looks something like this when developing:
    # bygg >>> bygg 0.8.5.dev6+g48b4c41.d20250504
    # Clean it and check if it seems valid. However, it will vary depending on what tags
    # are set etc, so only do a rudimentary check for numbers in a major-minor pattern.
    cleaned_version = process.stdout[14:].strip()
    assert re.match(r"\d+\.\d+", cleaned_version)


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_non_existing_action(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "no_such_action", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 1
    assert process.stdout == snapshot


def test_build_multiple_actions(snapshot, clean_bygg_tree):
    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / "taskrunner",
            "shorthand_action_yaml",
            "touch a file",
            "shorthand action yaml, with spaces",
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    cleaned_result = [
        line
        for line in process.stdout.split("\n")
        if "Starting process" not in line and "completed in" not in line
    ]
    assert "\n".join(sorted(cleaned_result)) == snapshot


def test_build_verbose(snapshot, clean_bygg_tree):
    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / "taskrunner",
            "hhgttg",
            "--verbose",
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    cleaned_result = [
        line for line in process.stdout.split("\n") if not line.startswith("bygg >>>")
    ]
    assert "\n".join(cleaned_result) == snapshot


def test_check(clean_bygg_tree):
    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / "checks",
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )

    assert process.returncode == 0

    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / "checks",
            "--check",
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert "The following checks reported issues:" in process.stdout


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_reset_remove_environments(
    snapshot, clean_bygg_tree, example: ExampleParameters
):
    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / example.name,
            *example.actions_for_all_environments,
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.build_rc

    for environment in example.environments:
        assert Path(
            clean_bygg_tree / examples_dir / example.name / environment
        ).exists()

    process = subprocess.run(
        ["bygg", "-C", examples_dir / example.name, "--remove-environments"],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot

    for environment in example.environments:
        assert not Path(
            clean_bygg_tree / examples_dir / example.name / environment
        ).exists()


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_reset_remove_cache(snapshot, clean_bygg_tree, example: ExampleParameters):
    process = subprocess.run(
        ["bygg", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.build_rc

    assert (
        Path(clean_bygg_tree / examples_dir / example.name / DEFAULT_DB_FILE).exists()
        == example.build_yields_cache
    )

    process = subprocess.run(
        ["bygg", "-C", examples_dir / example.name, "--remove-cache"],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot

    assert not Path(
        clean_bygg_tree / examples_dir / example.name / DEFAULT_DB_FILE
    ).exists()


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_reset_reset(snapshot, clean_bygg_tree, example: ExampleParameters):
    process = subprocess.run(
        [
            "bygg",
            "-C",
            examples_dir / example.name,
            *example.actions_for_all_environments,
        ],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.build_rc

    for environment in example.environments:
        assert Path(
            clean_bygg_tree / examples_dir / example.name / environment
        ).exists()

    assert (
        Path(clean_bygg_tree / examples_dir / example.name / DEFAULT_DB_FILE).exists()
        == example.build_yields_cache
    )

    process = subprocess.run(
        ["bygg", "-C", examples_dir / example.name, "--reset"],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot

    for environment in example.environments:
        assert not Path(
            clean_bygg_tree / examples_dir / example.name / environment
        ).exists()

    assert not Path(
        clean_bygg_tree / examples_dir / example.name / DEFAULT_DB_FILE
    ).exists()


# Rudimentary test to at least run the watch code
def test_watch():
    with pytest.raises(subprocess.TimeoutExpired):
        subprocess.run(
            ["bygg", "-C", examples_dir / "taskrunner", "succeed", "--watch"],
            capture_output=True,
            encoding="utf-8",
            timeout=2,
        )



import os
import time

from watchdog.events import DirModifiedEvent, FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer

from bygg.logutils import logger


class FileEventHandler(FileSystemEventHandler):
    def __init__(self, paths: set[str]):
        self.paths = paths
        self.has_matching_path = False

    def should_keep_running(self) -> bool:
        return True

    def on_modified(self, event: DirModifiedEvent | FileModifiedEvent):
        filename = os.path.realpath(event.src_path)
        if filename in self.paths:
            logger.debug("Modified event: %s", filename)
            self.has_matching_path = True


def do_watch(files_to_watch: set[str]):
    normalised_paths = set(os.path.realpath(path) for path in files_to_watch)
    logger.debug("Watching for file changes in: %s", normalised_paths)

    event_handler = FileEventHandler(normalised_paths)
    observer = Observer()
    observer.schedule(event_handler, ".", recursive=True)
    observer.start()

    try:
        while not event_handler.has_matching_path:
            time.sleep(1)
    finally:
        logger.debug("Stopping watcher")
        observer.stop()
        observer.join()



#!/usr/bin/env python3

from pathlib import Path

from bygg.output.output import TerminalStyle as TS
from bygg.output.output import output_error
from src.bygg.system_helpers import call

directories = [
    Path("src/bygg"),
    *{p.parent for p in Path("examples").glob("*/*.py")},
    Path("tests"),
]
mypy = Path(".venv/bin/mypy")
extra_arguments = {
    # We want to type check without having to have all the environments set up.
    Path("examples/environments/"): "--ignore-missing-imports",
}

exit_code = 0

for directory in directories:
    print(f"\n{TS.BOLD}=== Running mypy in {directory} ==={TS.RESET}")
    s = call(
        f"{mypy} --no-warn-no-return --disable-error-code func-returns-value --check-untyped-defs --pretty {extra_arguments.get(directory, '')} {directory}"
    )
    if not s:
        output_error(f"Failed to run mypy in {directory}")
        exit_code |= s

exit(exit_code)



import argparse
from dataclasses import dataclass, field
import shutil
import textwrap
from typing import Any, Literal, TypeAlias

from bygg.cmd.completions import ByggfileDirectoriesCompleter, EntrypointCompleter
from bygg.logutils import logger

MaintenanceCommand: TypeAlias = Literal["remove_cache", "remove_environments"]


@dataclass
class ByggNamespace:
    """Wrapper for the arguments parsed by argparse. Improves ergonimics when working
    with the arguments."""

    actions: list[str]
    version: bool
    verbose: bool
    is_restarted_with_env: str | None
    ipc_filename: list[str] | None
    clean: bool
    list_actions: bool
    tree: bool
    watch: bool
    jobs: int | None
    always_make: bool
    check: bool
    maintenance_commands: list[MaintenanceCommand]
    completions: bool
    dump_schema: bool
    directory: list[str] = field(default_factory=list)


class ByggHelpFormatter(argparse.RawDescriptionHelpFormatter):
    formatting_width = 80

    def __init__(self, prog):
        terminal_cols, terminal_rows = shutil.get_terminal_size()
        self.formatting_width = min(terminal_cols, self.formatting_width)

        indent_increment = 2
        max_help_position = 24

        super().__init__(
            prog, indent_increment, max_help_position, self.formatting_width
        )

    def _fill_text(self, text, width, indent):
        return fill_help_text(text, width, indent)


def fill_help_text(text: str, width: int, indent: str) -> str:
    return "\n".join(
        textwrap.fill(
            line,
            width=width,
            initial_indent=indent,
            subsequent_indent=indent + "  " if line.startswith("- ") else indent,
        )
        for line in text.strip().splitlines(keepends=True)
    )


def create_argument_parser():
    logger.info("Creating argument parser")

    parser = argparse.ArgumentParser(
        formatter_class=ByggHelpFormatter,
        description="""
A build tool written in Python, where all actions can be written in Python.

Build the default action:
 %(prog)s

Build ACTION:
 %(prog)s ACTION

Clean ACTION:
 %(prog)s ACTION --clean

List available actions:
 %(prog)s --list
""",
    )

    parser._positionals.title = "POSITIONAL ARGUMENTS"
    parser._optionals.title = "OPTIONS"

    # Use Any to get around type checking for argcomplete:

    arg: Any = parser.add_argument(
        "actions",
        nargs="*",
        default=None,
        help="Entrypoint actions to operate on.",
    )
    arg.completer = EntrypointCompleter()

    parser.add_argument(
        "-V", "--version", action="store_true", help="Show version string and exit."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable more verbose output from some commands.",
    )

    parser.add_argument(
        "--is_restarted_with_env",
        nargs="?",
        type=str,
        help=argparse.SUPPRESS,
    )
    # Used internally for communicating the IPC filename to the subprocess.
    parser.add_argument(
        "--ipc_filename",
        nargs=1,
        type=str,
        default=None,
        help=argparse.SUPPRESS,
    )
    # Commands that operate on the build setup:
    build_setup_wrapper_group = parser.add_argument_group(
        "Commands that operate on the build setup"
    )  # add_mutually_exclusive_group doesn't accept a title, so wrap it in a regular group.
    build_setup_group = build_setup_wrapper_group.add_mutually_exclusive_group()
    build_setup_group.add_argument(
        "--clean",
        action="store_true",
        help="Clean the outputs of the specified actions.",
    )
    build_setup_group.add_argument(
        "-l",
        "--list",
        action="store_true",
        dest="list_actions",
        help="List available actions.",
    )
    build_setup_group.add_argument(
        "--tree",
        action="store_true",
        help="Display the dependency tree starting from the specified action(s).",
    )
    build_setup_group.add_argument(
        "-w",
        "--watch",
        action="store_true",
        dest="watch",
        help="Watch the input files and rebuild when they change.",
    )
    # Some arguments inspired by Make:
    make_group = parser.add_argument_group("Make-like arguments")
    arg = make_group.add_argument(
        "-C",
        "--directory",
        nargs=1,
        type=str,
        default=None,
        help="Change to the specified directory.",
    )
    arg.completer = ByggfileDirectoriesCompleter()

    make_group.add_argument(
        "-j",
        "--jobs",
        nargs="?",
        type=int,
        default=None,
        help="Specify the number of jobs to run simultaneously. None means to use the number of available cores.",
    )
    make_group.add_argument(
        "-B",
        "--always-make",
        action="store_true",
        help="Always build all actions.",
    )

    # Analyse and verify:
    analyse_group = parser.add_argument_group(
        "Analyse and verify",
        """
Arguments in this group will add more analysis to the build process. Actions will be built and the analysis result will be reported.

There are also two environment variables available:
- BYGG_DEBUG enables debug logging.
- BYGG_DEV treats all actions as entry points, which allows for running them directly from the command line.
""",
    )
    analyse_group.add_argument(
        "--check",
        action="store_true",
        help="Perform various checks on the action tree. Implies -B",
    )

    # Maintenance arguments:
    maintenance_group = parser.add_argument_group(
        "Maintenance",
        "These commands operate on Bygg's build state. When given a maintenance command, Bygg will just perform the maintenance and exit; it will neither build anything nor perform any other actions.",
    )
    maintenance_group.add_argument(
        "--reset",
        action=AppendConstList,
        const=["remove_cache", "remove_environments"],
        nargs=0,
        dest="maintenance_commands",
        help="Remove the cache and the Python environments. Same as giving --remove-cache and --remove-environments.",
    )
    maintenance_group.add_argument(
        "--remove-cache",
        action="append_const",
        const="remove_cache",
        dest="maintenance_commands",
        help="Remove the build cache.",
    )
    maintenance_group.add_argument(
        "--remove-environments",
        action="append_const",
        const="remove_environments",
        dest="maintenance_commands",
        help="Remove the Python environments.",
    )

    # Meta arguments:
    meta_group = parser.add_argument_group("Meta arguments")
    meta_group.add_argument(
        "--dump-schema",
        action="store_true",
        help="Generate a JSON Schema for the Byggfile.yml files. The schema will be printed to stdout.",
    )
    meta_group.add_argument(
        "--completions",
        action="store_true",
        help="Output instructions for how to set up shell completions via the shell's startup script.",
    )

    return parser


class AppendConstList(argparse.Action):
    """Custom action to append the elements from a list to dest."""

    def __call__(self, parser, namespace, values, option_string=None):
        const_list = getattr(namespace, self.dest) or list()
        const_list.extend(self.const)
        setattr(namespace, self.dest, const_list)



import datetime
import logging
import os

from bygg.output.output import TerminalStyle as TS


def __getattr__(name):
    if name == "logger":
        return logging.getLogger("bygg")
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


class CustomLogFormatter(logging.Formatter):
    COLORS = {
        "DEBUG": TS.Fg.BLUE,
        "INFO": TS.Fg.GREEN,
        "WARNING": TS.Fg.YELLOW,
        "ERROR": TS.Fg.RED,
        "CRITICAL": TS.Fg.RED + TS.BOLD,
    }

    no_colors = False

    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        no_colors: bool = False,
    ):
        super().__init__(fmt, datefmt)
        self.no_colors = no_colors

    def format(self, record: logging.LogRecord) -> str:
        # Doesn't care about the format string used when instantiating the logger
        level_color = self.COLORS.get(record.levelname, "")
        filename_width = 40
        # Add the length of the colour code so the width will be the same for no_colors
        color_filename_width = filename_width + len(TS.Fg.WHITE)

        filename = (
            f"{record.filename + ':' + str(record.lineno):<{filename_width}}"
            if self.no_colors
            else f"{TS.Fg.CYAN}{record.filename + ':' + TS.Fg.WHITE + str(record.lineno):<{color_filename_width}}{TS.RESET}"
        )

        message = record.getMessage()

        if self.no_colors:
            return f"{self.formatTime(record)} | {record.name} | {record.levelname:<8} | {filename} | {message}"

        return f"{self.formatTime(record)} | {TS.Fg.MAGENTA}{record.name}{TS.RESET} | {level_color}{record.levelname:<8}{TS.RESET} | {filename} | {message}"

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:
        formatted_time = super().formatTime(record, datefmt)
        if self.no_colors:
            return formatted_time
        return f"{TS.Fg.CYAN}{formatted_time}{TS.RESET}"


def setup_logging():
    debug = str.lower(os.environ.get("BYGG_DEBUG", ""))

    if not debug:
        logging.getLogger("bygg").disabled = True
        return

    logger = logging.getLogger("bygg")
    logger.setLevel(logging.DEBUG)

    if "silent" not in debug:
        colored_formatter = CustomLogFormatter()

        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.DEBUG)
        stream_handler.setFormatter(colored_formatter)
        logger.addHandler(stream_handler)

    # file or silent enables file debugging
    if "file" in debug or "silent" in debug:
        bw_formatter = CustomLogFormatter(no_colors=True)

        iso_datetime = datetime.datetime.now().isoformat(timespec="seconds")
        file_handler = logging.FileHandler(f"debug_{iso_datetime}.log")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(bw_formatter)
        logger.addHandler(file_handler)

        # Create symlink to latest log
        latest_log_link = "debug_latest.log"
        if os.path.lexists(latest_log_link):
            os.unlink(latest_log_link)
        os.symlink(f"debug_{iso_datetime}.log", latest_log_link)

    logger.debug("Debug logging enabled")



from dataclasses import dataclass
import os
import re
import subprocess
from typing import Optional

from bygg.core.common_types import CommandStatus


def create_shell_command(shell_command: str, message: Optional[str] = None):
    def call_shell_command(inputs: Optional[set[str]], outputs: Optional[set[str]]):
        # Map stderr onto stdout and return both as the output
        process = subprocess.run(
            shell_command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
        )
        return CommandStatus(process.returncode, message, process.stdout)

    return lambda ctx: call_shell_command(ctx.inputs, ctx.outputs)


@dataclass
class FileListsFromPattern:
    input_files: list[str]
    output_files: list[str]
    unmatched_input_files: list[str]


def filenames_from_pattern(
    input_files: list[str], in_pattern: str, target_pattern: str
) -> FileListsFromPattern:
    """
    Generate a list of target filenames from a list of input file names and a pattern
    like in Makefile pattern rules. File names that don't match the pattern are ignored
    and returned as a separate list.

    Returns a FileListsFromPattern dataclass containing lists of the input and output
    filenames as well as the ignored input files.

    Example filename transformation:

      ["/foo/bar/baz.txt.j2"], "%.txt.j2", "out_%.txt")
        => "/foo/bar/out_baz.txt"

    Intended to work like the corresponding mechanism in GNU Makefiles. From
    https://www.gnu.org/software/make/manual/html_node/Pattern-Match.html:

    "A target pattern is composed of a ‘%’ between a prefix and a suffix, either or both
    of which may be empty. The pattern matches a file name only if the file name starts
    with the prefix and ends with the suffix, without overlap. The text between the
    prefix and the suffix is called the stem. Thus, when the pattern ‘%.o’ matches the
    file name test.o, the stem is ‘test’. The pattern rule prerequisites are turned into
    actual file names by substituting the stem for the character ‘%’. Thus, if in the
    same example one of the prerequisites is written as ‘%.c’, it expands to ‘test.c’.

    When the target pattern does not contain a slash (and it usually does not),
    directory names in the file names are removed from the file name before it is
    compared with the target prefix and suffix. After the comparison of the file name to
    the target pattern, the directory names, along with the slash that ends them, are
    added on to the prerequisite file names generated from the pattern rule’s
    prerequisite patterns and the file name. The directories are ignored only for the
    purpose of finding an implicit rule to use, not in the application of that rule.
    Thus, ‘e%t’ matches the file name src/eat, with ‘src/a’ as the stem. When
    prerequisites are turned into file names, the directories from the stem are added at
    the front, while the rest of the stem is substituted for the ‘%’. The stem ‘src/a’
    with a prerequisite pattern ‘c%r’ gives the file name src/car."
    """

    matched_input_files = []
    output_files = []
    unmatched_input_files = []
    regex = f"^{in_pattern.replace('%', '(.+)')}$"
    matcher = re.compile(regex)

    for input_file in input_files:
        input_file_head, input_file_tail = os.path.split(input_file)
        target_pattern_head, _ = os.path.split(target_pattern)

        outfilename = ""

        if target_pattern_head:
            # Target pattern has a directory, so compare them including directories
            match_object = matcher.match(input_file)
            if match_object:
                stem = match_object.group(1)
                outfilename = target_pattern.replace("%", stem)
        else:
            # Target pattern does not have a directory, so strip each filename from the
            # directory part before comparing
            match_object = matcher.match(input_file_tail)
            if match_object:
                stem = match_object.group(1)
                outfilename = os.path.join(
                    input_file_head, target_pattern.replace("%", stem)
                )

        if outfilename:
            matched_input_files.append(input_file)
            output_files.append(outfilename)
        else:
            unmatched_input_files.append(input_file)

    return FileListsFromPattern(
        matched_input_files,
        output_files,
        unmatched_input_files,
    )



from dataclasses import dataclass
import re

from bygg.core.scheduler import Job
from bygg.output.output import TerminalStyle as TS
from bygg.output.output import isatty, output_plain


@dataclass
class HighlightConfig:
    pattern: str | re.Pattern[str]
    start: str
    end: str


warning_hl_config = HighlightConfig(
    pattern=r"(warn\w*:?|WARN\w*:?)",
    start=TS.Bg.YELLOW + TS.Fg.BLACK,
    end=TS.Fg.RESET + TS.Bg.RESET,
)

error_hl_config = HighlightConfig(
    pattern=r"(error\w*:?|ERR!)",
    start=TS.Bg.RED,
    end=TS.Fg.RESET + TS.Bg.RESET,
)

default_highlight_config: list[HighlightConfig] = [warning_hl_config, error_hl_config]


def highlight_log(message: str, config: list[HighlightConfig] | None = None):
    """
    Highlights the message according to the regexes and styles. Regexes are applied
    per line.
    """
    if not config:
        return message

    lines = []
    for line in message.splitlines():
        for c in config:
            line = re.sub(c.pattern, c.start + r"\1" + c.end, line, flags=re.IGNORECASE)
        lines.append(line)
    return "\n".join(lines)


def output_job_logs(jobs: list[Job]):
    print(
        f"\n{TS.BOLD}Showing logs for {len(jobs)} failed job{'s' if len(jobs) > 1 else ''}:{TS.RESET}\n"
    )
    for job in jobs:
        if job.status and (log := job.status.output):
            output_plain(f'{TS.BOLD}--- Start "{job.name}" ---{TS.RESET}')
            output_plain(
                highlight_log(log, default_highlight_config) if isatty else log
            )
            output_plain(f'{TS.BOLD}--- End "{job.name}" ---{TS.RESET}')
            output_plain("")


def format_job_log(job: Job) -> str:
    log = job.status.output if job.status else None
    formatted_log = []

    if not log:
        formatted_log.append(f"{TS.BOLD}No output from job {job.name}.{TS.RESET}")
        return ""
    formatted_log.append(f'{TS.BOLD}--- Start "{job.name}" ---{TS.RESET}')
    formatted_log.append(
        highlight_log(log, default_highlight_config) if isatty else log
    )
    formatted_log.append(f'{TS.BOLD}--- End "{job.name}" ---{TS.RESET}')
    formatted_log.append("")

    return "\n".join(formatted_log)



import argparse
import functools


@functools.cache
def _build_dest_to_action(parser: argparse.ArgumentParser) -> dict:
    actions = parser._get_optional_actions() + parser._get_positional_actions()
    dest_to_action = {action.dest: action for action in actions}
    return dest_to_action


def _get_argument_for_dest(parser: argparse.ArgumentParser, dest: str) -> str | None:
    dest_to_action: dict[str, argparse.Action] = _build_dest_to_action(parser)
    action = dest_to_action.get(dest, None)
    if action is None:
        return None
    option_strings = action.option_strings
    return option_strings[-1] if option_strings else ""


def unparse_args(
    parser: argparse.ArgumentParser,
    args: argparse.Namespace,
    *,
    drop: list[str] | None = None,
) -> list[str]:
    """
    Convert an argparse.Namespace back to a list of command line arguments.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser that was used to parse the command line arguments.
    args : argparse.Namespace
        The parsed command line arguments to unparse.
    drop : list[str] | None, optional
        Which dest keys to drop, by default None.

    Returns
    -------
    list[str]
        A list of command line arguments
    """
    exec_list: list[str] = []
    for k, v in vars(args).items():
        argument = _get_argument_for_dest(parser, k)
        if drop and k in drop:
            continue
        if v is False or v is None:
            # These are arguments that were not given
            continue
        if v is True:
            # If argument is None, something wrong is broken in the argparse Namespace
            # and should be looked into.
            assert argument
            exec_list.append(argument)
        elif v:
            if argument:
                exec_list.append(
                    f"{argument}={','.join(v) if isinstance(v, (list, tuple)) else v}"
                )
            elif isinstance(v, (list, tuple)):
                exec_list.extend(v)
            else:
                assert False
        elif argument == "":
            continue
        else:
            # Could happen if we add another type of argument
            assert False
    return exec_list



import sys

isatty = sys.stdout.isatty()


class TerminalStyle:
    """Terminal Text Styling"""

    CLEARLINE = "\033[2K\r" if isatty else ""
    RESET = "\033[0m" if isatty else ""
    BOLD = "\033[1m" if isatty else ""
    NOBOLD = "\033[22m" if isatty else ""
    DIM = "\033[2m" if isatty else ""
    UNDERLINE = "\033[4m" if isatty else ""
    BLINK = "\033[5m" if isatty else ""
    INVERTED = "\033[7m" if isatty else ""
    STRIKETHROUGH = "\033[9m" if isatty else ""
    HIDE_CURSOR = "\033[?25l" if isatty else ""
    SHOW_CURSOR = "\033[?25h" if isatty else ""

    class Fg:
        """Foreground Text Color"""

        BLACK = "\033[30m" if isatty else ""
        RED = "\033[31m" if isatty else ""
        GREEN = "\033[32m" if isatty else ""
        YELLOW = "\033[33m" if isatty else ""
        BLUE = "\033[34m" if isatty else ""
        MAGENTA = "\033[35m" if isatty else ""
        CYAN = "\033[36m" if isatty else ""
        WHITE = "\033[37m" if isatty else ""
        RESET = "\033[39m" if isatty else ""

        BRIGHT_RED = "\033[91m" if isatty else ""
        BRIGHT_GREEN = "\033[92m" if isatty else ""
        BRIGHT_YELLOW = "\033[93m" if isatty else ""
        BRIGHT_BLUE = "\033[94m" if isatty else ""
        BRIGHT_MAGENTA = "\033[95m" if isatty else ""
        BRIGHT_CYAN = "\033[96m" if isatty else ""
        BRIGHT_WHITE = "\033[97m" if isatty else ""

    class Bg:
        """Background Text Color"""

        BLACK = "\033[40m" if isatty else ""
        RED = "\033[41m" if isatty else ""
        GREEN = "\033[42m" if isatty else ""
        YELLOW = "\033[43m" if isatty else ""
        BLUE = "\033[44m" if isatty else ""
        MAGENTA = "\033[45m" if isatty else ""
        CYAN = "\033[46m" if isatty else ""
        WHITE = "\033[47m" if isatty else ""
        RESET = "\033[49m" if isatty else ""

        BRIGHT_RED = "\033[101m" if isatty else ""
        BRIGHT_GREEN = "\033[102m" if isatty else ""
        BRIGHT_YELLOW = "\033[103m" if isatty else ""
        BRIGHT_BLUE = "\033[104m" if isatty else ""
        BRIGHT_MAGENTA = "\033[105m" if isatty else ""
        BRIGHT_CYAN = "\033[106m" if isatty else ""
        BRIGHT_WHITE = "\033[107m" if isatty else ""


def output_with_status_line(bottom: str | None, scroll: str | None):
    """
    Outputs a line of text at the bottom of the terminal which is cleared and reprinted,
    and another line of text that scrolls up.

    Only prints the latter if the terminal is not a tty.
    """
    if not isatty:
        print(scroll)
        return

    print(TerminalStyle.CLEARLINE, end="")

    if scroll is not None:
        print(scroll)
    print(bottom if bottom is not None else "", end="\r")


STATUS_TEXT_FIELD_WIDTH = 8


bygg_prefix_string = f"{TerminalStyle.Fg.BRIGHT_CYAN}{'bygg >>>':<{STATUS_TEXT_FIELD_WIDTH}}{TerminalStyle.Fg.RESET}{TerminalStyle.NOBOLD}"


def output_info(s: str):
    print(
        f"{bygg_prefix_string} {TerminalStyle.Fg.BRIGHT_CYAN}{s}{TerminalStyle.Fg.RESET}"
    )


def output_warning(s: str):
    print(
        f"{bygg_prefix_string} {TerminalStyle.Fg.BRIGHT_YELLOW}{s}{TerminalStyle.Fg.RESET}"
    )


def output_error(s: str):
    print(
        f"{bygg_prefix_string} {TerminalStyle.Fg.BRIGHT_RED}{s}{TerminalStyle.Fg.RESET}"
    )


def output_ok(s: str):
    print(
        f"{bygg_prefix_string} {TerminalStyle.Fg.BRIGHT_GREEN}{s}{TerminalStyle.Fg.RESET}"
    )


def output_plain(s: str):
    print(s)



from bygg.core.action import Action
from bygg.core.common_types import CommandStatus


class Job:
    name: str
    action: Action
    status: CommandStatus | None

    def __init__(self, action: Action):
        self.name = action.name
        self.action = action
        self.status = None

    def __repr__(self) -> str:
        return f'"{self.name}, status: {self.status.rc if self.status else "unknown"}"'

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Job) and self.name == __o.name



import dis
import functools
import hashlib
import io
import os
from pathlib import Path
from typing import Callable

DIGEST_TYPE = "sha1"

# If True, cache the file digest in memory based on the filename, stat.ctime_ns,
# stat.mtime_ns and stat.size of the file.
ALLOW_DIGEST_CACHING = True


def file_digest(file: str | Path) -> str:
    with open(file, "rb") as f:
        return hashlib.file_digest(f, DIGEST_TYPE).hexdigest()


@functools.cache
def file_digest_memo(
    file: str, st_ctime_ns: int, st_mtime_ns: int, st_size: int
) -> str:
    """
    Cache a file's digest based on the content, stat.ctime_ns, stat.mtime_ns and
    stat.size of the file.
    """
    return file_digest(file)


def calculate_file_digest(file: str | Path) -> str | None:
    """
    Calculate the digest of a file.

    file: The file to calculate the digest of.
    Returns: The digest of the file as a hex string, or None if the file does not exist.

    """
    real_path = os.path.realpath(file)
    if os.path.isfile(real_path):
        if ALLOW_DIGEST_CACHING:
            st = os.stat(real_path)
            return file_digest_memo(
                real_path, st.st_ctime_ns, st.st_mtime_ns, st.st_size
            )
        return file_digest(real_path)

    if os.path.isdir(real_path):
        return "directory"


def calculate_dependency_digest(filenames: set[str] | set[Path]) -> tuple[str, bool]:
    """
    Calculate the digest of a set of files.
    filenames: The files to calculate the digest of.
    Returns: The digest of the files as a hex string.
    """

    file_digests = [calculate_file_digest(filename) for filename in filenames]
    digests = sorted(filter(None, file_digests))
    files_were_missing = len(file_digests) != len(digests)

    return (
        hashlib.new(DIGEST_TYPE, "".join(digests).encode()).hexdigest(),
        files_were_missing,
    )


def calculate_digest(items: list[str]) -> str:
    digests = sorted(
        filter(
            None,
            [hashlib.new(DIGEST_TYPE, item.encode()).hexdigest() for item in items],
        )
    )

    return hashlib.new(DIGEST_TYPE, "".join(digests).encode()).hexdigest()


def calculate_function_digest(fn: Callable) -> str:
    """
    Calculate the digest of a function.
    fn: The function to calculate the digest of.
    Returns: The digest of the function as a hex string.
    """
    # Just looking at the code object is not enough, so let's disassemble it.
    out = io.StringIO()
    dis.dis(fn, file=out)
    function_dis_string = out.getvalue()
    return hashlib.new(DIGEST_TYPE, function_dis_string.encode()).hexdigest()


def calculate_string_digest(s: str) -> str:
    """
    Calculate the digest of a string.
    s: The string to calculate the digest of.
    Returns: The digest of the string as a hex string.
    """
    return hashlib.new(DIGEST_TYPE, s.encode()).hexdigest()



import rich

from bygg.core.action import ActionContext, action
from bygg.core.common_types import CommandStatus


@action("action1_python")
def example_action(ctx: ActionContext):
    rich.print("[magenta]Running default action from Python.[/magenta]")
    return CommandStatus(0, "All ok.", None)


@action("action1_python_entrypoint_python", is_entrypoint=True)
def example_action_entrypoint(ctx: ActionContext):
    """An entrypoint action declared in Python."""
    rich.print("[magenta]Running entrypoint action from Python.[/magenta]")
    return CommandStatus(0, "All ok.", None)



"""
A build tool written in Python, where all actions can be written in Python.
"""

# PYTHON_ARGCOMPLETE_OK

from bygg.cmd.dispatcher import bygg
from bygg.logutils import logger, setup_logging
from bygg.output.output import (
    TerminalStyle as TS,
)
from bygg.output.output import output_warning


def main():
    print(TS.HIDE_CURSOR, end="")
    setup_logging()
    logger.info("Starting")
    try:
        return bygg()
    except KeyboardInterrupt:
        output_warning("Interrupted by user. Aborting.")
        return 1
    finally:
        print(TS.SHOW_CURSOR, end="")


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod
from collections import deque
from typing import Any, Iterable

from bygg.core.action import Action


class Dag(metaclass=ABCMeta):
    def __len__(self) -> int:
        return 0

    def clear(self): ...

    def remove_node(self, node: str): ...

    def build_action_graph(self, build_actions: dict[str, Action], action: Action): ...

    @abstractmethod
    def get_ready_jobs(
        self,
        finished_jobs: dict[str, Any],
        running_jobs: dict[str, Any],
    ) -> list[str]: ...

    @abstractmethod
    def get_all_jobs(self) -> Iterable[str]: ...


class ByggDag(Dag):
    nodes: dict[str, set[str]]

    def __init__(self):
        self.nodes = {}

    def __len__(self) -> int:
        return len(self.nodes)

    def clear(self):
        self.nodes.clear()

    def remove_node(self, node: str):
        self.nodes.pop(node, None)

    def build_action_graph(self, build_actions: dict[str, Action], action: Action):
        """Build the action graph."""
        queue = deque([action])
        while len(queue) > 0:
            a = queue.popleft()
            self.nodes[a.name] = set()
            for dependency in a.dependencies:
                dependency_action = build_actions.get(dependency)
                if not dependency_action:
                    raise ValueError(f"Action '{dependency}' not found")
                queue.append(dependency_action)
                self.nodes[a.name].add(dependency)

    def get_ready_jobs(
        self,
        finished_jobs: dict[str, Any],
        running_jobs: dict[str, Any],
    ) -> list[str]:
        nodes = set(self.nodes.keys())
        ready_jobs = []

        for node, dependencies in self.nodes.items():
            # Successful and skipped jobs will have been removed from the graph, while
            # failed jobs remain.
            if node in finished_jobs or node in running_jobs:
                continue
            # The job is ready to run if its dependencies are no longer in the graph.
            if len(nodes & dependencies) == 0:
                ready_jobs.append(node)
        return ready_jobs

    def get_all_jobs(self) -> Iterable[str]:
        return self.nodes.keys()


def create_dag() -> Dag:
    return ByggDag()



#!/usr/bin/env python3

from pathlib import Path
import shutil
import sys

from src.bygg.system_helpers import call

venv = Path(".venv")

uv_bin = shutil.which("uv")
venv_command = (
    "python3 -m venv"
    if not uv_bin
    # Leave uv out of options but to use the Python active in the shell:
    else f"{uv_bin} venv --no-project --no-managed-python"
)
pip_install_command = (
    ".venv/bin/pip install"
    if not uv_bin
    else f"{uv_bin} pip install --reinstall --no-progress"
)

# Always remove existing venv, to make sure requirements are correctly installed.
if venv.exists():
    print(f"Removing existing {venv} directory")
    shutil.rmtree(venv)

print("Creating Python virtualenv")
s = call(f"{venv_command} {venv}")
if not s or not venv.exists():
    print("Failed to create Python virtualenv")
    sys.exit(s)

print("Installing requirements:")
s = call(f"{pip_install_command} -r requirements.txt -r requirements-dev.txt")
if not s:
    print("Failed to install requirements")
    sys.exit(s)

print("Installing myself:")
s = call(f"{pip_install_command} -e .")
if not s:
    print("Failed to install Bygg")
    sys.exit(s)



import os
from pathlib import Path
import signal
import sys
from typing import Callable
import warnings

from bygg.core.action import CommandStatus
from bygg.core.common_types import JobStatus
from bygg.core.scheduler import Job, Scheduler
from bygg.output.output import TerminalStyle as TS
from bygg.output.status_display import on_check_failed

JobStatusListener = Callable[[JobStatus, Job, tuple], None]
RunnerStatusListener = Callable[[str], None]


# Suppress the specific loky warning about fork start method. The current runner
# architecture depends on forking, and at least from what I can discern from reading in
# the loky source code, this is only a problem on Windows.
warnings.filterwarnings(
    "ignore",
    message="`fork` start method should not be used with `loky`",
    category=UserWarning,
)


class ProcessRunner:
    scheduler: Scheduler
    job_status_listener: JobStatusListener
    runner_status_listener: RunnerStatusListener
    failed_jobs: list[Job]

    def __init__(self, scheduler: Scheduler):
        self.scheduler = scheduler
        self.job_status_listener = lambda *args: None
        self.runner_status_listener = lambda *args: None
        self.failed_jobs = []

    def start(self, max_workers: int = 1) -> list[Job]:
        from loky import Future, ProcessPoolExecutor, wait  # type: ignore
        from loky.backend import get_context  # type: ignore

        total_job_count = len(self.scheduler.job_graph)

        self.runner_status_listener(
            f"Starting process runner with {max_workers} threads"
        )

        def init_worker():
            """Ignore CTRL+C in the worker process."""
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        with ProcessPoolExecutor(
            max_workers=max_workers,
            initializer=init_worker,
            context=get_context("fork"),
        ) as pool:
            scheduled_jobs: dict[Job, Future] = {}
            backlog: list[Job] = []

            def call_status_listener():
                for job in scheduled_jobs.keys():
                    self.job_status_listener(
                        "running",
                        job,
                        (len(self.scheduler.finished_jobs), total_job_count),
                    )

            def get_job_count_tuple():
                return (
                    len(self.scheduler.finished_jobs) + len(self.failed_jobs),
                    total_job_count,
                )

            # If a job fails, we want to stop scheduling new jobs and just wait for the
            # ones that are already running to finish.
            exit_reasons: list[Job] = []

            while True:
                if (
                    not exit_reasons
                    and len(backlog) < 2 * max_workers
                    and (jobs := self.scheduler.get_ready_jobs())
                ):
                    backlog += jobs

                if (
                    len(scheduled_jobs) == 0
                    and len(backlog) == 0
                    and (self.scheduler.run_status() == "finished" or exit_reasons)
                ):
                    return exit_reasons

                # Keep the scheduled queue relatively short; no need to schedule much
                # more than we have workers
                batch_size = max_workers * 2
                if len(scheduled_jobs) < batch_size:
                    for job in backlog:
                        # Skip jobs with no command or ones that are clean
                        if job.action.command is None:
                            job.status = CommandStatus(
                                0,
                                f"{'No command, skipping'}",
                                None,
                            )
                            self.scheduler.job_finished(job)
                            self.job_status_listener(
                                "skipped",
                                job,
                                get_job_count_tuple(),
                            )
                            backlog.remove(job)
                            continue

                        # Run in-process
                        if job.action.scheduling_type == "in-process":
                            self.job_status_listener(
                                "running",
                                job,
                                get_job_count_tuple(),
                            )
                            job.status = job.action.command(job.action)
                            self.scheduler.job_finished(job)
                            self.job_status_listener(
                                "finished" if job.status.rc == 0 else "failed",
                                job,
                                get_job_count_tuple(),
                            )
                            backlog.remove(job)
                            continue

                        # Schedule job to be run on the worker processes
                        backlog.remove(job)
                        scheduled_jobs[job] = pool.submit(run_job, (job))

                call_status_listener()

                if len(scheduled_jobs) == 0:
                    continue

                completed_jobs, _ = wait(
                    scheduled_jobs.values(), timeout=0.1, return_when="FIRST_COMPLETED"
                )
                for future in completed_jobs:
                    job_result = future.result()
                    assert isinstance(job_result, Job)
                    del scheduled_jobs[job_result]

                    self.check_for_missing_output_files(job_result)

                    self.scheduler.job_finished(job_result)
                    if job_result.status is not None and job_result.status.rc == 0:
                        if job_result.status.runner_instruction:
                            exit_reasons.append(job_result)
                        self.job_status_listener(
                            "finished",
                            job_result,
                            get_job_count_tuple(),
                        )
                    else:
                        self.failed_jobs.append(job_result)
                        self.job_status_listener(
                            "failed",
                            job_result,
                            get_job_count_tuple(),
                        )
                        call_status_listener()
                        exit_reasons.append(job_result)

    def check_for_missing_output_files(self, job: Job):
        missing_files: list[str | Path] = []
        for filename in job.action.outputs:
            if not os.path.exists(filename):
                missing_files.append(filename)
        if missing_files:
            on_check_failed(
                "output_file_missing",
                job.action,
                f"Job {TS.BOLD}{job.name}{TS.RESET} didn't create the output file{'s' if len(missing_files) > 1 else ''} that it declared: {TS.BOLD}{', '.join([str(missing_file) for missing_file in missing_files])}{TS.RESET}",
                "error",
            )


def run_job(job: Job):
    try:
        if job.action.command is None:
            job.status = CommandStatus(0, "No command, skipping", None)
        else:
            job.status = job.action.command(job.action)
    except Exception as e:
        job.status = CommandStatus(1, "Job failed with exception.", f"{e}")
    return job


def get_job_count_limit():
    if sys.version_info >= (3, 13):
        return os.process_cpu_count() or 1
    try:
        # Use os.sched_getaffinity where available (on U**X):
        # https://stackoverflow.com/a/55423170
        return len(os.sched_getaffinity(0))
    except AttributeError:
        count = os.cpu_count()
        assert count is not None
        return count
